from __future__ import annotations

import textwrap
from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.plugins.os.unix.linux._os import LinuxPlugin
from dissect.target.plugins.os.unix.linux.network_managers import NetworkManager
from dissect.target.tools.query import main as target_query
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("expected_ips", "messages"),
    [
        (
            ["10.13.37.1"],
            "Jan  1 13:37:01 hostname NetworkManager[1]: <info>  [1600000000.0000] dhcp4 (eth0): option ip_address           => '10.13.37.1'",  # noqa: E501
        ),
        (["10.13.37.2"], "Feb  2 13:37:02 test systemd-networkd[2]: eth0: DHCPv4 address 10.13.37.2/24 via 10.13.37.0"),
        (
            ["10.13.37.3"],
            "Mar  3 13:37:03 localhost NetworkManager[3]: <info>  [1600000000.0003] dhcp4 (eth0):   address 10.13.37.3",
        ),
        (
            ["10.13.37.4"],
            "Apr  4 13:37:04 localhost dhclient[4]: bound to 10.13.37.4 -- renewal in 1337 seconds.",
        ),
        (
            ["2001:db8::"],
            (
                "Jun  6 13:37:06 test systemd-networkd[5]: eth0: DHCPv6 address 2001:db8::/64 via 2001:db8:ffff:ffff:ffff:ffff:ffff:ffff\n"  # noqa: E501
                "May  5 13:37:05 test systemd-networkd[5]: eth0: DHCPv6 lease lost\n"
            ),
        ),
    ],
)
def test_ips_dhcp(
    target_unix_users: Target, fs_unix: VirtualFilesystem, expected_ips: list[str], messages: str
) -> None:
    """Test DHCP lease messages from /var/log/syslog."""

    fs_unix.map_file_fh(
        "/var/log/syslog",
        BytesIO(textwrap.dedent(messages).encode()),
    )

    target_unix_users.add_plugin(LinuxPlugin)
    results = target_unix_users.ips
    results.reverse()
    assert len(results) == len(expected_ips)
    assert sorted(results) == expected_ips


@pytest.mark.parametrize(
    ("flag", "expected_out"),
    [
        (None, "['10.13.37.2']"),
        # ("--dhcp-all", "['10.13.37.2', '10.13.37.1']"),
        # Temporarily disabled behaviour, for discussion see:
        # https://github.com/fox-it/dissect.target/pull/687#discussion_r1698515269
    ],
)
def test_ips_dhcp_arg(
    target_unix: Target,
    fs_unix: VirtualFilesystem,
    flag: str,
    expected_out: str,
    capsys: pytest.CaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test --dhcp-all flag behaviour."""

    fs_unix.map_file_fh("/etc/timezone", BytesIO(b"Europe/Amsterdam"))

    messages = """
    Apr  1 13:37:01 localhost dhclient[4]: bound to 10.13.37.1 -- renewal in 1337 seconds.
    Apr  2 13:37:02 localhost foo[1]: some other message.
    Apr  3 13:37:03 localhost dhclient[4]: bound to 10.13.37.2 -- renewal in 1337 seconds.
    """

    fs_unix.map_file_fh(
        "/var/log/syslog",
        BytesIO(textwrap.dedent(messages).encode()),
    )
    target_unix.add_plugin(LinuxPlugin)

    argv = ["target-query", "foo", "-f", "ips"]
    if flag:
        argv.append(flag)

    with patch("dissect.target.Target.open_all", return_value=[target_unix]), monkeypatch.context() as m:
        m.setattr("sys.argv", argv)
        target_query()
        out, _ = capsys.readouterr()
        assert expected_out in out


def test_ips_cloud_init(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test cloud-init dhcp.py lease messages."""

    messages = """
    2022-12-31 13:37:00,000 - dhcp.py[DEBUG]: Received dhcp lease on eth0 for 10.13.37.5/24
    """

    fs_unix.map_file_fh(
        "/var/log/cloud-init.log",
        BytesIO(textwrap.dedent(messages).encode()),
    )

    target_unix_users.add_plugin(LinuxPlugin)
    results = target_unix_users.ips

    assert len(results) == 1
    assert results == ["10.13.37.5"]


def test_ips_static(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test statically defined ipv4 and ipv6 addresses in /etc/network/interfaces."""

    fs_unix.map_file("/etc/network/interfaces", absolute_path("_data/plugins/os/unix/_os/ips/interfaces"))
    target_unix_users.add_plugin(LinuxPlugin)
    results = target_unix_users.ips

    assert sorted(results) == sorted(["10.13.37.6", "2001:db8:ffff:ffff:ffff:ffff:ffff:ffff"])


def test_ips_wicked_static(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test statically defined ipv4 addresses in /etc/wicked/ifconfig/."""

    fs_unix.map_file("/etc/wicked/ifconfig/eth0.xml", absolute_path("_data/plugins/os/unix/_os/ips/eth0.xml"))
    target_unix_users.add_plugin(LinuxPlugin)
    results = target_unix_users.ips

    assert sorted(results) == sorted(["10.13.37.2", "2001:db8:ffff:ffff:ffff:ffff:ffff:fffe"])


def test_dns_static(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test statically defined ipv4 and ipv6 dns-nameservers in /etc/network/interfaces."""

    fs_unix.map_file("/etc/network/interfaces", absolute_path("_data/plugins/os/unix/_os/ips/interfaces"))
    target_unix_users.add_plugin(LinuxPlugin)
    results = target_unix_users.dns

    assert results == [{"10.13.37.1", "10.13.37.2", "2001:db8::", "2001:db9::"}]


def test_ips_netplan_static(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test statically defined ipv4 and ipv6 ip addresses in /etc/netplan/*.yaml."""

    config = """
    # This file describes the network interfaces available on your system
    # For more information, see netplan(5).
    network:
        version: 2
        renderer: networkd
        ethernets:
            enp0s3:
                dhcp4: no
                addresses: [192.168.1.123/24]
                gateway4: 192.168.1.1
                nameservers:
                    addresses: [1.2.3.4, 5.6.7.8]
    """

    fs_unix.map_file_fh("/etc/netplan/01-netcfg.yaml", BytesIO(textwrap.dedent(config).encode()))
    target_unix_users.add_plugin(LinuxPlugin)
    assert target_unix_users.ips == ["192.168.1.123"]


@pytest.mark.parametrize(
    ("config", "expected_output"),
    [
        ("", []),
        ("network:", []),
        ("network:\n    ethernets:\n", []),
        ("network:\n    ethernets:\n        eth0:\n", []),
        ("network:\n    ethernets:\n        eth0:\n            addresses: []\n", []),
        ("network:\n    ethernets:\n        eth0:\n            addresses: [1.2.3.4/24]\n", ["1.2.3.4"]),
        ("network:\n    ethernets:\n        eth0:\n            addresses: ['1.2.3.4']\n", ["1.2.3.4"]),
    ],
)
def test_ips_netplan_static_invalid(
    target_unix_users: Target, fs_unix: VirtualFilesystem, config: str, expected_output: list
) -> None:
    fs_unix.map_file_fh("/etc/netplan/02-netcfg.yaml", BytesIO(textwrap.dedent(config).encode()))
    target_unix_users.add_plugin(LinuxPlugin)
    assert target_unix_users.ips == expected_output


def test_ips_netplan_static_empty_regression(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/netplan/01-netcfg.yaml", BytesIO())
    target_unix_users.add_plugin(LinuxPlugin)
    assert target_unix_users.ips == []


@pytest.mark.parametrize(
    ("input", "expected_output"),
    [
        # 'invalid' or input that should be filtered
        ("0.0.0.0", set()),
        ("127.0.0.1", set()),
        ("127.0.0.1/8", set()),
        ("0.0.0.0/24", set()),
        ("::1", set()),
        ("::", set()),
        ("0:0:0:0:0:0:0:1", set()),
        # valid input
        ("::ffff:192.0.2.128", {"::ffff:192.0.2.128"}),
        ("2001:db8::2:1", {"2001:db8::2:1"}),
        ("10.13.37.1", {"10.13.37.1"}),
        ("10.13.37.2/24", {"10.13.37.2"}),
        ("  10.13.37.3  ", {"10.13.37.3"}),
        ("2001:db8::", {"2001:db8::"}),
        ("2001:db8:ffff:ffff:ffff:ffff:ffff:ffff", {"2001:db8:ffff:ffff:ffff:ffff:ffff:ffff"}),
    ],
)
def test_clean_ips(input: str, expected_output: set) -> None:
    """Test the cleaning of dirty ip addresses."""

    assert NetworkManager.clean_ips({input}) == expected_output


def test_regression_ips_unique_strings(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Regression test for https://github.com/fox-it/dissect.target/issues/877."""

    config = """
    network:
        ethernets:
            eth0:
                addresses: ['1.2.3.4']
    """
    fs_unix.map_file_fh("/etc/netplan/01-netcfg.yaml", BytesIO(textwrap.dedent(config).encode()))
    fs_unix.map_file_fh("/etc/netplan/02-netcfg.yaml", BytesIO(textwrap.dedent(config).encode()))

    syslog = "Apr  4 13:37:04 localhost dhclient[4]: bound to 1.2.3.4 -- renewal in 1337 seconds."
    fs_unix.map_file_fh("/var/log/syslog", BytesIO(textwrap.dedent(syslog).encode()))

    target_unix.add_plugin(LinuxPlugin)

    assert isinstance(target_unix.ips, list)
    assert all(isinstance(ip, str) for ip in target_unix.ips)

    assert len(target_unix.ips) == 1
    assert target_unix.ips == ["1.2.3.4"]


def test_ips_dhcp_lease_files(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we can detect DHCP lease files from NetworkManager and dhclient."""

    lease1 = """
    # This is private data. Do not parse.
    ADDRESS=1.2.3.4
    """

    lease2 = """
    lease {
        interface "eth0";
        fixed-address 9.0.1.2;
        option dhcp-lease-time 13337;
        option routers 0.0.0.0;
        option host-name "hostname";
        renew 1 2023/12/31 13:37:00;
        rebind 2 2023/01/01 01:00:00;
        expire 3 2024/01/01 13:37:00;
        # real leases contain more key/value pairs
    }
    lease {
        interface "eth0";
        fixed-address 5.6.7.8;
        option dhcp-lease-time 13337;
        option routers 0.0.0.0;
        option host-name "hostname";
        renew 1 2024/12/31 13:37:00;
        rebind 2 2024/01/01 01:00:00;
        expire 3 2025/01/01 13:37:00;
        # real leases contain more key/value pairs
    }
    """

    lease3 = """
    some-other "value";
    lease {
        interface "eth1";
        fixed-address 3.4.5.6;
    }
    """

    fs_unix.map_file_fh("/var/lib/NetworkManager/internal-uuid-eth0.lease", BytesIO(textwrap.dedent(lease1).encode()))
    fs_unix.map_file_fh("/var/lib/dhcp/dhclient.leases", BytesIO(textwrap.dedent(lease2).encode()))
    fs_unix.map_file_fh("/var/lib/dhclient/dhclient.eth0.leases", BytesIO(textwrap.dedent(lease3).encode()))

    syslog = "Apr  4 13:37:04 localhost dhclient[4]: bound to 1.3.3.7 -- renewal in 1337 seconds."
    fs_unix.map_file_fh("/var/log/syslog", BytesIO(textwrap.dedent(syslog).encode()))

    target_unix.add_plugin(LinuxPlugin)

    # tests if we did not call :func:`parse_unix_dhcp_log_messages` since :func:`parse_unix_dhcp_leases` has results.
    assert len(target_unix.ips) == 4
    assert sorted(target_unix.ips) == sorted(["1.2.3.4", "5.6.7.8", "9.0.1.2", "3.4.5.6"])



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugin import OperatingSystem
from dissect.target.plugins.os.unix.bsd.freebsd._os import FreeBsdPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_bsd_freebsd_os_detection(target_bare: Target) -> None:
    """Test if we detect FreeBSD correctly."""

    fs = VirtualFilesystem()
    fs.makedirs("/net")
    fs.map_file("/bin/freebsd-version", absolute_path("_data/plugins/os/unix/bsd/freebsd/freebsd-freebsd-version"))

    target_bare.filesystems.add(fs)
    target_bare.apply()

    assert FreeBsdPlugin.detect(target_bare)
    assert isinstance(target_bare._os, FreeBsdPlugin)
    assert target_bare.os == OperatingSystem.BSD



from __future__ import annotations

from dissect.target.plugin import InternalNamespacePlugin


class KeyProviderPlugin(InternalNamespacePlugin):
    """Key provider plugin for DPAPI."""

    __namespace__ = "dpapi.keyprovider"

    def check_compatible(self) -> None:
        return None



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from flow.record.fieldtypes import datetime as dt

from dissect.target.plugins.apps.av.symantec import (
    SEPFirewallRecord,
    SEPLogRecord,
    SymantecPlugin,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_symantec_firewall(target_win: Target, fs_win: VirtualFilesystem) -> None:
    log_file = absolute_path("_data/plugins/apps/av/symantec/tralog.log")
    fs_win.map_file("ProgramData/Symantec/Symantec Endpoint Protection/Today/Data/Logs/tralog.log", log_file)
    target_win.add_plugin(SymantecPlugin)
    records = list(target_win.symantec.firewall())
    assert len(records) == 3
    assert isinstance(records[0], type(SEPFirewallRecord()))
    assert records[0].ts == dt("2023-08-18T13:44:28.219290Z")
    assert str(records[0].local_ip) == "10.0.2.15"
    assert str(records[0].remote_ip) == "10.0.2.2"
    assert bool(records[0].outbound) is False
    assert records[0].local_port == 25
    assert records[0].remote_port == 49510
    assert records[0].user == "who"
    assert records[0].repetition == 4
    assert bool(records[0].blocked) is True
    assert records[0].severity == "Info"
    assert records[0].rule_id == 0
    assert records[0].remote_host == ""
    assert records[0].rule_name == "Block all IP traffic"
    assert records[0].application == Path()
    assert records[0].line_no == 2


def test_symantec_log(target_win: Target, fs_win: VirtualFilesystem) -> None:
    log_file = absolute_path("_data/plugins/apps/av/symantec/cve.log")
    fs_win.map_file("ProgramData/Symantec/Symantec Endpoint Protection/Today/Data/Logs/av/cve.log", log_file)
    target_win.add_plugin(SymantecPlugin)
    records = list(target_win.symantec.logs())
    assert len(records) == 1
    assert isinstance(records[0], type(SEPLogRecord()))
    assert records[0].ts == dt("2023-08-18T15:18:46Z")
    assert records[0].virus == "EICAR Test String"
    assert records[0].source_file == Path("C:\\eicar.com")
    assert records[0].action_taken == "Pending Analysis"
    assert records[0].virus_type == "Test"
    assert records[0].scan_id == "1692364523"
    assert records[0].quarantine_id == "0"
    assert records[0].virus_id == 11101
    assert records[0].depth == 0
    assert bool(records[0].still_infected) is False
    assert bool(records[0].quarantined) is False
    assert bool(records[0].compressed) is False
    assert bool(records[0].cleanable) is False
    assert bool(records[0].deletable) is False
    assert records[0].confidence == 0
    assert records[0].prevalence == 0
    assert records[0].risk == 0
    assert records[0].download_url == ""
    assert records[0].line_no == 1



from __future__ import annotations

import io
import re
import string
from typing import TYPE_CHECKING, BinaryIO, Callable, Union

if TYPE_CHECKING:
    import logging
    from collections.abc import Iterator

    from dissect.target.helpers.record import TargetRecordDescriptor


Needle = Union[bytes, re.Pattern]


def find_needles(
    fh: BinaryIO,
    needles: Needle | list[Needle],
    *,
    start: int | None = None,
    end: int | None = None,
    lock_seek: bool = True,
    block_size: int = io.DEFAULT_BUFFER_SIZE,
    progress: Callable[[int], None] | None = None,
) -> Iterator[tuple[Needle, int, re.Match | None]]:
    """Yields needles, their offsets and optional regex match found in provided byte stream.

    Args:
        fh: The byte stream to search for needles.
        needles: The list of bytes needles to search for.
        start: The offset to start searching from.
        end: The offset to stop searching at.
        lock_seek: Whether the file position is maintained by the scraper or the consumer.
                   Setting this to ``False`` will allow the consumer to seek the file pointer, i.e. to skip forward.
        block_size: The block size to use for reading from the byte stream.
        progress: A function to call with the current offset.
    """

    if not isinstance(needles, list):
        needles = [needles]

    if not needles:
        raise ValueError("At least one needle value must be provided")

    if start is not None and end is not None and start >= end:
        raise ValueError("Start offset must be less than end offset")

    max_needle_len = max(len(n.pattern if isinstance(n, re.Pattern) else n) for n in needles)
    overlap_len = max_needle_len

    offset = fh.tell() if start is None else start
    current_block = b""

    while offset < end if end is not None else True:
        if lock_seek:
            fh.seek(offset)
        else:
            offset = fh.tell()

        read_size = min(block_size, end - offset) if end is not None else block_size
        if not (next_block := fh.read(read_size)):
            break

        if progress:
            progress(offset)

        overlap = current_block[-overlap_len:] if overlap_len > 0 else b""
        current_block = overlap + next_block
        current_block_offset = offset - len(overlap)

        # Look for a needle in a current block
        last_needle_pos = -1
        last_needle_end = 0
        while True:
            needle_positions = []

            for needle in needles:
                match = None
                if (
                    pos := (match.start(0) if (match := needle.search(current_block, last_needle_pos + 1)) else -1)
                    if isinstance(needle, re.Pattern)
                    else current_block.find(needle, last_needle_pos + 1)
                ) > -1:
                    needle_positions.append((pos, needle, match))

            if not needle_positions:
                break

            closest_needle_pos, closest_needle, match = min(needle_positions)
            yield closest_needle, current_block_offset + closest_needle_pos, match

            last_needle_pos = closest_needle_pos
            last_needle_end = last_needle_pos + 1

        # The size of the data from the current block that will be prepended to the next block
        overlap_len = min(len(current_block) - last_needle_end, max_needle_len)

        offset += block_size


def _read_plain_chunk(fh: BinaryIO, needle: Needle, offset: int, chunk_size: int) -> bytes:
    """Read chunk from a byte stream for provided needle, offset and chunk size."""
    fh.seek(offset)
    return fh.read(chunk_size)


def find_needle_chunks(
    fh: BinaryIO,
    needle_chunk_size_map: dict[Needle, int],
    chunk_reader: Callable[[BinaryIO, Needle, int, int], bytes] | None = None,
    lock_seek: bool = True,
    block_size: int = io.DEFAULT_BUFFER_SIZE,
) -> Iterator[tuple[bytes, int, bytes, re.Match | None]]:
    """Yields tuples with an offset, a needle, a byte chunk and optional regex match found in provided byte stream.

    Args:
        fh: The byte stream to search for needles.
        needle_chunk_size_map: A dictionary with needle bytes as keys and chunk sizes as values.
        chunk_reader: A function to read a chunk from a byte stream for provided needle, offset and chunk size.
        lock_seek: Whether the file position is maintained by the scraper or the consumer.
                   Setting this to ``False`` wil allow the consumer to seek the file pointer, i.e. to skip forward.
        block_size: The block size to use for reading from the byte stream.
    """
    chunk_reader = chunk_reader or _read_plain_chunk

    needles = list(needle_chunk_size_map.keys())
    for needle, offset, match in find_needles(fh, needles, lock_seek=lock_seek, block_size=block_size):
        yield (needle, offset, chunk_reader(fh, needle, offset, needle_chunk_size_map[needle]), match)


def scrape_chunks(
    fh: BinaryIO,
    needle_chunk_size_map: dict[Needle, int],
    chunk_parser: Callable[[Needle, bytes], Iterator[TargetRecordDescriptor]],
    chunk_reader: Callable[[BinaryIO, Needle, int, int], bytes] | None = None,
    block_size: int = io.DEFAULT_BUFFER_SIZE,
    log: logging.Logger | None = None,
) -> Iterator[TargetRecordDescriptor]:
    """Yields records scraped from chunks found in a provided byte stream.

    Args:
        fh: The byte stream to search for needles.
        needle_chunk_size_map: A dictionary with needle bytes as keys and chunk sizes as values.
        chunk_parser: A function to parse a chunk and yield records.
        chunk_reader: A function to read a chunk from a byte stream for provided needle, offset and chunk size.
        block_size: The block size to use for reading from the byte stream.
        log: A logger to use for logging.
    """

    chunk_count = 0
    record_count = 0

    chunk_stream = find_needle_chunks(
        fh,
        needle_chunk_size_map,
        block_size=block_size,
        chunk_reader=chunk_reader or _read_plain_chunk,
    )

    for needle, _, chunk, _ in chunk_stream:
        try:
            for record in chunk_parser(needle, chunk):
                yield record
                record_count += 1
            chunk_count += 1
        except Exception as e:  # noqa: PERF203
            if log:
                log.warning("Chunk parsing failed with %r", e)

    if log:
        log.debug("Scraped %d chunks with %d records", chunk_count, record_count)


def recover_string(buf: bytes, encoding: str, *, reverse: bool = False, ascii: bool = True) -> str:
    """Recover the longest possible string from a byte buffer, forward or reverse.

    Args:
        buf: The byte buffer to recover a string from.
        encoding: The encoding to use for decoding the buffer.
        reverse: Whether to recover the string from the end of the buffer.
        ascii: Whether to recover only ASCII characters.
    """
    if reverse:
        decoded = ""
        error_count = 0
        max_error_count = 8  # Arbitrary limit

        for i in range(1, len(buf) + 1):
            try:
                decoded = buf[-i:].decode(encoding)
                error_count = 0
            except UnicodeDecodeError:  # noqa: PERF203
                error_count += 1

                if error_count > max_error_count:
                    break

        decoded = decoded.split("\x00")[-1]

        if ascii:
            for i in range(1, len(decoded) + 1):
                if decoded[-i] not in string.printable:
                    decoded = decoded[-i + 1 :]
                    break
    else:
        try:
            decoded = buf.decode(encoding)
        except UnicodeDecodeError as e:
            decoded = buf[: e.start].decode(encoding)

        decoded = decoded.split("\x00")[0]

        if ascii:
            for i in range(len(decoded)):
                if decoded[i] not in string.printable:
                    decoded = decoded[:i]
                    break

    return decoded



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

ProcProcessRecord = TargetRecordDescriptor(
    "linux/proc/processes",
    [
        ("datetime", "ts"),
        ("string", "name"),
        ("string", "state"),
        ("varint", "pid"),
        ("datetime", "runtime"),
        ("varint", "ppid"),
        ("string", "parent"),
    ],
)


class ProcProcesses(Plugin):
    """Linux ``/proc`` process volatile plugin."""

    def check_compatible(self) -> None:
        if not self.target.has_function("proc"):
            raise UnsupportedPluginError("proc filesystem not available")

    @export(record=ProcProcessRecord)
    def processes(self) -> Iterator[ProcProcessRecord]:
        """Return the processes available in ``/proc`` and the stats associated with them.

        There is a numerical subdirectory for each running process; the subdirectory is named by the process ID.
        Each ``/proc/[pid]`` subdirectory contains various pseudo-files.

        Yields ProcProcessRecord with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The start time of the process.
            name (string): The name of the process.
            state (string): The state of the process.
            pid (int): The process ID of the process.
            runtime (datetime): The amount of time the process is running until moment of acquisition.
            ppid (int): The parent process ID of the process.
            parent (string): The name of the parent process ID.
        """

        for process in self.target.proc.processes():
            yield ProcProcessRecord(
                ts=process.starttime,
                name=process.name,
                pid=process.pid,
                runtime=process.runtime.seconds,
                ppid=process.ppid,
                state=process.state,
                parent=process.parent.name,
                _target=self.target,
            )



from __future__ import annotations

import socket
import ssl
from pathlib import Path
from struct import unpack
from typing import Callable
from unittest.mock import call, patch

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.remote import RemoteLoader, RemoteStream, RemoteStreamConnection
from dissect.target.target import Target


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``RemoteLoader`` when opening a ``Target``."""
    path = "remote://127.0.0.1:1337"
    with (
        patch.object(ssl, "SSLContext", autospec=True),
        patch("dissect.target.loaders.remote.RemoteLoader.map"),
        patch("dissect.target.target.Target.apply"),
    ):
        target = opener(path)
        assert isinstance(target._loader, RemoteLoader)
        assert target.path == Path("127.0.0.1:1337")


def test_loader() -> None:
    """Test that ``RemoteLoader`` is correctly selected."""
    with patch.object(ssl, "SSLContext", autospec=True):
        loader = loader_open("remote://127.0.0.1:1337")
        assert isinstance(loader, RemoteLoader)


def test_stream() -> None:
    """Test that we can create a ``RemoteStream`` with a ``RemoteStreamConnection``."""

    with (
        patch.object(ssl, "SSLContext", autospec=True) as mock_ssl_context,
        patch.object(socket, "socket", autospec=True) as mock_socket,
    ):
        rsc = RemoteStreamConnection("remote://127.0.0.1", 9001, options={"ca": "A", "key": "B", "crt": "C"})
        assert rsc.is_connected() is False
        rsc.connect()

        rsc.remote_disk_data = b"ABC"
        rsc.remote_disk_response = None

        def send(data: bytes) -> None:
            _, offset, read = unpack(">BQQ", data)
            rsc.remote_disk_response = rsc.remote_disk_data[offset : offset + read]
            return len(data)

        def receive(num: bytes) -> bytes:
            return rsc.remote_disk_response

        rsc._ssl_sock.send = send
        rsc._ssl_sock.recv = receive
        rs = RemoteStream(rsc, 0, 3)
        rs.align = 1
        rs.seek(1)
        data = rs.read(2)
        rs.close()

        mock_socket.assert_called_with(socket.AddressFamily.AF_INET, socket.SocketKind.SOCK_STREAM)

        expected = [
            call(ssl.PROTOCOL_TLSv1_2),
            call().load_default_certs(),
            call().load_cert_chain(certfile="C", keyfile="B"),
            call().load_verify_locations("A"),
            call().wrap_socket(rsc._socket, server_hostname="remote://127.0.0.1"),
            call().wrap_socket().connect(("remote://127.0.0.1", 9001)),
        ]

        assert data == b"BC"
        assert mock_ssl_context.mock_calls == expected
        assert rs.tell() == 3
        assert rsc.is_connected() is True


def test_stream_embedded() -> None:
    with (
        patch.object(ssl, "SSLContext", autospec=False) as mock_ssl_context,
        patch.object(socket, "socket", autospec=True),
    ):
        RemoteStreamConnection.configure("K", "C")
        RemoteStreamConnection("remote://127.0.0.1", 9001)
        mock_ssl_context.assert_has_calls([call().load_cert_chain_str(certfile="C", keyfile="K")])



from __future__ import annotations

import os
import sys
import textwrap
from functools import reduce
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest
from docutils.core import publish_string
from docutils.utils import SystemMessage
from flow.record import RecordDescriptor

from dissect.target.exceptions import PluginError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import EmptyRecord, create_extended_descriptor
from dissect.target.plugin import (
    FunctionDescriptor,
    FunctionDescriptorLookup,
    InternalNamespacePlugin,
    InternalPlugin,
    NamespacePlugin,
    OSPlugin,
    Plugin,
    PluginDescriptor,
    PluginDescriptorLookup,
    PluginRegistry,
    _find_py_files,
    _save_plugin_import_failure,
    alias,
    environment_variable_paths,
    export,
    find_functions,
    find_functions_by_record_field_type,
    functions,
    get_external_module_paths,
    load_modules_from_paths,
    lookup,
    plugins,
)
from dissect.target.plugins.apps.other.env import EnvironmentFilePlugin
from dissect.target.plugins.general.users import UsersPlugin
from dissect.target.plugins.os.default._os import DefaultOSPlugin
from dissect.target.plugins.os.unix.linux._os import LinuxPlugin
from dissect.target.plugins.os.unix.linux.debian._os import DebianPlugin
from dissect.target.plugins.os.unix.linux.fortios._os import FortiOSPlugin
from dissect.target.plugins.os.windows._os import WindowsPlugin
from dissect.target.target import Target

if TYPE_CHECKING:
    from collections.abc import Iterator

    from flow.record import Record
    from pytest_benchmark.fixture import BenchmarkFixture


def test_save_plugin_import_failure() -> None:
    test_trace = ["test-trace"]
    test_module_name = "test-module"

    with (
        patch("traceback.format_exception", Mock(return_value=test_trace)),
        patch("dissect.target.plugin.PLUGINS", new_callable=PluginRegistry) as mock_plugins,
    ):
        _save_plugin_import_failure(test_module_name)

        assert len(mock_plugins.__failed__) == 1
        assert mock_plugins.__failed__[0].module == test_module_name
        assert mock_plugins.__failed__[0].stacktrace == test_trace


@pytest.mark.parametrize(
    ("env_value", "expected_output"),
    [
        (None, []),
        ("", []),
        (":", [Path(), Path()]),
    ],
)
def test_load_environment_variable(env_value: str | None, expected_output: list[Path]) -> None:
    with patch.object(os, "environ", {"DISSECT_PLUGINS": env_value}):
        assert environment_variable_paths() == expected_output


def test_load_module_paths() -> None:
    assert get_external_module_paths([Path(), Path()]) == [Path()]


def test_load_paths_with_env() -> None:
    with patch.object(os, "environ", {"DISSECT_PLUGINS": ":"}):
        assert get_external_module_paths([Path(), Path()]) == [Path()]


def test_load_environment_variable_empty_string() -> None:
    with patch("dissect.target.plugin._find_py_files") as mocked_find_py_files:
        load_modules_from_paths([])
        mocked_find_py_files.assert_not_called()


def test_load_environment_variable_comma_seperated_string() -> None:
    with patch("dissect.target.plugin._find_py_files") as mocked_find_py_files:
        load_modules_from_paths([Path(), Path()])
        mocked_find_py_files.assert_called_with(Path())


def test_filter_file(tmp_path: Path) -> None:
    file = tmp_path / "hello.py"
    file.touch()

    assert list(_find_py_files(file)) == [file]

    test_file = tmp_path / "non_existent_file"
    assert list(_find_py_files(test_file)) == []

    test_file = tmp_path / "__init__.py"
    test_file.touch()
    assert list(_find_py_files(test_file)) == []


@pytest.mark.parametrize(
    ("filename", "empty_list"),
    [
        ("__init__.py", True),
        ("__pycache__/help.pyc", True),
        ("hello/test.py", False),
    ],
)
def test_filter_directory(tmp_path: Path, filename: str, empty_list: bool) -> None:
    file = tmp_path / filename
    file.parent.mkdir(parents=True, exist_ok=True)
    file.touch()

    if empty_list:
        assert list(_find_py_files(tmp_path)) == []
    else:
        assert file in list(_find_py_files(tmp_path))


@pytest.mark.parametrize(
    ("filename", "expected_module"),
    [
        ("test.py", "test"),
        ("hello_world/help.py", "hello_world.help"),
        ("path/to/file.py", "path.to.file"),
    ],
)
def test_filesystem_module_registration(tmp_path: Path, filename: str, expected_module: str) -> None:
    path = tmp_path / filename
    path.parent.mkdir(parents=True, exist_ok=True)
    path.touch()

    load_modules_from_paths([tmp_path])

    assert expected_module in sys.modules


def test_plugin_registration(tmp_path: Path) -> None:
    code = """
        from dissect.target.plugin import Plugin, export


        class TestPlugin(Plugin):
            __register__ = False

            def check_compatible(self) -> None:
                return None

            @export(output="default")
            def hello_world(self):
                for x in self.target.fs.iterdir(""):
                    print(f"hello {x}")
    """
    (tmp_path / "plugin.py").write_text(textwrap.dedent(code))

    with patch("dissect.target.plugin.register") as mock_register:
        load_modules_from_paths([tmp_path])

        mock_register.assert_called_once()
        assert mock_register.call_args[0][0].__name__ == "TestPlugin"

    with patch("dissect.target.plugin.register") as mock_register:
        load_modules_from_paths([tmp_path / "plugin.py"])

        mock_register.assert_called_once()
        assert mock_register.call_args[0][0].__name__ == "TestPlugin"


@patch("dissect.target.plugin.PLUGINS", new_callable=PluginRegistry)
def test_plugin_directory(mock_plugins: PluginRegistry, tmp_path: Path) -> None:
    code = """
        from dissect.target.plugin import Plugin, export

        class MyPlugin(Plugin):
            __namespace__ = {!r}

            @export
            def my_function(self):
                return "My function"
    """

    (tmp_path / "myplugin").mkdir()
    (tmp_path / "myplugin" / "__init__.py").write_text("")
    (tmp_path / "myplugin" / "_plugin.py").write_text(textwrap.dedent(code.format(None)))

    (tmp_path / "mypluginns").mkdir()
    (tmp_path / "mypluginns" / "__init__.py").write_text("")
    (tmp_path / "mypluginns" / "_plugin.py").write_text(textwrap.dedent(code.format("myns")))

    load_modules_from_paths([tmp_path])

    assert mock_plugins.__functions__.__regular__ == {
        "my_function": {
            "myplugin.MyPlugin": FunctionDescriptor(
                name="my_function",
                namespace=None,
                path="myplugin.my_function",
                exported=True,
                internal=False,
                findable=True,
                alias=False,
                output="default",
                method_name="my_function",
                module="myplugin._plugin",
                qualname="MyPlugin",
            )
        },
        "myns": {
            "mypluginns.MyPlugin": FunctionDescriptor(
                name="myns",
                namespace="myns",
                path="mypluginns",
                exported=True,
                internal=False,
                findable=True,
                alias=False,
                output="record",
                method_name="__call__",
                module="mypluginns._plugin",
                qualname="MyPlugin",
            )
        },
        "myns.my_function": {
            "mypluginns.MyPlugin": FunctionDescriptor(
                name="myns.my_function",
                namespace="myns",
                path="mypluginns.my_function",
                exported=True,
                internal=False,
                findable=True,
                alias=False,
                output="default",
                method_name="my_function",
                module="mypluginns._plugin",
                qualname="MyPlugin",
            )
        },
    }

    assert mock_plugins.__plugins__.__regular__ == {
        "myplugin.MyPlugin": PluginDescriptor(
            module="myplugin._plugin",
            qualname="MyPlugin",
            namespace=None,
            path="myplugin",
            findable=True,
            functions=["my_function"],
            exports=["my_function"],
        ),
        "mypluginns.MyPlugin": PluginDescriptor(
            module="mypluginns._plugin",
            qualname="MyPlugin",
            namespace="myns",
            path="mypluginns",
            findable=True,
            functions=["my_function", "__call__"],
            exports=["my_function", "__call__"],
        ),
    }


TestRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "application/test",
    [
        ("string", "value"),
    ],
)

TestRecord2 = create_extended_descriptor([UserRecordDescriptorExtension])(
    "application/test_other",
    [
        ("varint", "value"),
    ],
)


class _TestNSPlugin(NamespacePlugin):
    __namespace__ = "NS"
    __register__ = False

    @export(record=TestRecord)
    def test_all(self) -> Iterator[TestRecord]:
        # Iterate all functions of all subclasses
        yield from self.test()


class _TestSubPlugin1(_TestNSPlugin):
    __namespace__ = "t1"
    __register__ = False

    @export(record=TestRecord)
    def test(self) -> Iterator[TestRecord]:
        yield TestRecord(value="test1")


class _TestSubPlugin2(_TestNSPlugin):
    __namespace__ = "t2"
    __register__ = False

    @export(record=TestRecord)
    def test(self) -> Iterator[TestRecord]:
        yield TestRecord(value="test2")


class _TestSubPlugin3(_TestSubPlugin2):
    __namespace__ = "t3"
    __register__ = False

    # Override the test() function of t2
    @export(record=TestRecord)
    def test(self) -> Iterator[TestRecord]:
        yield TestRecord(value=self._value())

    def _value(self) -> str:
        return "test3"


class _TestSubPlugin4(_TestSubPlugin3):
    __namespace__ = "t4"
    __register__ = False

    # Do not override the test() function of t3, but change the _value function instead.
    def _value(self) -> str:
        return "test4"

    @export(record=TestRecord)
    def test_other(self) -> Iterator[TestRecord]:
        yield TestRecord(value="test4-other")

    @export(record=TestRecord)
    def test_all(self) -> Iterator[TestRecord]:
        yield TestRecord(value="overridden")


class _TestSubPlugin5(_TestNSPlugin):
    __namespace__ = "t5"
    __register__ = False

    @export(record=TestRecord2)
    def test_other(self) -> Iterator[TestRecord2]:
        yield TestRecord2(value=69)


def test_namespace_plugin(target_win: Target) -> None:
    assert "__subplugins__" in dir(_TestNSPlugin)
    # Rename the test functions to protect them from being filtered by NS

    target_win._register_plugin_functions(_TestSubPlugin1(target_win))
    target_win._register_plugin_functions(_TestSubPlugin2(target_win))
    target_win._register_plugin_functions(_TestSubPlugin3(target_win))
    target_win._register_plugin_functions(_TestSubPlugin4(target_win))
    target_win._register_plugin_functions(_TestSubPlugin5(target_win))
    target_win._register_plugin_functions(_TestNSPlugin(target_win))

    assert len(target_win.NS.__subplugins__) == 5
    assert len(target_win.NS.test.__subplugins__) == 4
    assert target_win.NS.test.__doc__ == "Return test for: t1, t2, t3, t4"
    assert [rd.name for rd in target_win.NS.test.__record__] == ["application/test"]

    assert target_win.NS.test_other.__doc__ == "Return test_other for: t4, t5"
    assert sorted([rd.name for rd in target_win.NS.test_other.__record__]) == [
        "application/test",
        "application/test_other",
    ]
    assert len(target_win.NS.test_other.__subplugins__) == 2

    assert len(list(target_win.NS.test())) == 4
    assert sorted([item.value for item in target_win.NS.test()]) == ["test1", "test2", "test3", "test4"]
    assert sorted([item.value for item in target_win.t1.test()]) == ["test1"]
    assert sorted([item.value for item in target_win.t2.test()]) == ["test2"]
    assert sorted([item.value for item in target_win.t3.test()]) == ["test3"]
    assert sorted([item.value for item in target_win.t4.test()]) == ["test4"]

    # Check whether we can access all subclass functions from the superclass
    assert sorted([item.value for item in target_win.NS.test_all()]) == ["test1", "test2", "test3", "test4"]

    # Check whether we can access the overridden function when explicitly accessing the subplugin
    assert next(target_win.t4.test_all()).value == "overridden"

    with pytest.raises(PluginError, match="Cannot merge namespace methods with different output types"):

        class _TestSubPluginFaulty(_TestNSPlugin):
            __namespace__ = "faulty"
            __register__ = False

            @export(output="yield")
            def test(self) -> Iterator[str]:
                yield "faulty"


@patch("dissect.target.plugin.PLUGINS", new_callable=PluginRegistry)
def test_namespace_plugin_registration(mock_plugins: PluginRegistry) -> None:
    class _TestNSPlugin(NamespacePlugin):
        __namespace__ = "NS"

    class _TestSubPlugin1(_TestNSPlugin):
        __namespace__ = "t1"

        @export(record=TestRecord)
        def test(self) -> None: ...

    assert next(lookup("NS")).exported
    assert next(lookup("NS.test")).exported
    assert next(lookup("t1")).exported
    assert next(lookup("t1.test")).exported


@patch("dissect.target.plugin.PLUGINS", new_callable=PluginRegistry)
def test_namesplace_plugin_multiple_same_module(mock_plugins: PluginRegistry) -> None:
    class NS(NamespacePlugin):
        __namespace__ = "ns"

        def check_compatible(self) -> None:
            return None

    class Foo(NS):
        __namespace__ = "foo"

        @export(output="yield")
        def baz(self) -> Iterator[str]:
            yield from ["foo"]

    class Bar(NS):
        __namespace__ = "bar"

        @export(output="yield")
        def baz(self) -> Iterator[str]:
            yield from ["bar"]

    result, _ = find_functions("*.baz")
    assert len(result) == 2
    assert sorted(desc.name for desc in result) == ["bar.baz", "foo.baz"]


@patch("dissect.target.plugin.PLUGINS", new_callable=PluginRegistry)
def test_nested_namespace_getattr(mock_plugins: PluginRegistry, target_bare: Target) -> None:
    class Foo(Plugin):
        __namespace__ = "foo"

        @export(output="yield")
        def buzz(self) -> Iterator[str]:
            yield from ["buzz"]

    class FooBar(Plugin):
        __namespace__ = "foo.bar"

        @export(output="yield")
        def bazz(self) -> Iterator[str]:
            yield from ["bazz1"]

        @export(output="yield")
        def bar(self) -> Iterator[str]:
            yield from ["bar1"]

    class FooBaz(Plugin):
        __namespace__ = "foo.baz"

        @export(output="yield")
        def bazz(self) -> Iterator[str]:
            yield from ["bazz2"]

        @export(output="yield")
        def bar(self) -> Iterator[str]:
            yield from ["bar2"]

    for plugin in [Foo, FooBar, FooBaz]:
        target_bare._register_plugin_functions(plugin(target_bare))

    assert isinstance(target_bare.foo, Foo)
    assert isinstance(target_bare.foo.bar, FooBar)
    assert isinstance(target_bare.foo.baz, FooBaz)
    assert hasattr(target_bare.foo.bar, "bazz")

    with pytest.raises(AttributeError):
        target_bare.foo.bazz()

    with pytest.raises(AttributeError):
        target_bare.foo.bar.foo()

    # Test whether we can access the plugin this way
    assert next(target_bare.foo.bar.bazz()) == "bazz1"
    assert next(target_bare.foo.bar.bar()) == "bar1"
    assert next(target_bare.foo.baz.bazz()) == "bazz2"
    assert next(target_bare.foo.baz.bar()) == "bar2"
    assert next(target_bare.foo.buzz()) == "buzz"


def test_find_plugin_function_default(target_default: Target) -> None:
    found, _ = find_functions("services", target_default)

    assert len(found) == 2
    names = [item.name for item in found]
    assert "services" in names
    assert "services" in names
    paths = [item.path for item in found]
    assert "os.unix.linux.services.services" in paths
    assert "os.windows.services.services" in paths

    found, _ = find_functions("mcafee.msc", target_default)
    assert len(found) == 1
    assert found[0].path == "apps.av.mcafee.msc"

    found, _ = find_functions("webserver.access", target_default)
    assert len(found) == 1
    assert found[0].path == "apps.webserver.webserver.access"


@pytest.mark.parametrize(
    "pattern",
    [
        ("version,ips,hostname"),
        ("ips,version,hostname"),
        ("hostname,ips,version"),
        ("users,osinfo"),
        ("osinfo,users"),
    ],
)
def test_find_plugin_function_order(target_win: Target, pattern: str) -> None:
    found = ",".join(reduce(lambda rs, el: [*rs, el.method_name], find_functions(pattern, target_win)[0], []))
    assert found == pattern


class _TestIncompatiblePlugin(Plugin):
    def check_compatible(self) -> None:
        raise UnsupportedPluginError("My incompatible plugin error")


def test_incompatible_plugin(target_bare: Target) -> None:
    with pytest.raises(UnsupportedPluginError, match="My incompatible plugin error"):
        target_bare.add_plugin(_TestIncompatiblePlugin)


# A realistic slimmed down mock of a PluginRegistry
MOCK_PLUGINS = PluginRegistry(
    __plugins__=PluginDescriptorLookup(
        __regular__={
            # Non-OS, regular
            "general.osinfo.OSInfoPlugin": PluginDescriptor(
                module="dissect.target.plugins.general.osinfo",
                qualname="OSInfoPlugin",
                namespace=None,
                path="general.osinfo",
                findable=True,
                functions=["osinfo"],
                exports=["osinfo"],
            ),
            # Non-OS, with namespace
            "apps.chat.chat.ChatPlugin": PluginDescriptor(
                module="dissect.target.plugins.apps.chat.chat",
                qualname="ChatPlugin",
                namespace="chat",
                path="apps.chat.chat",
                findable=False,
                functions=["history", "__call__"],
                exports=["history", "__call__"],
            ),
            "apps.chat.msn.MSNPlugin": PluginDescriptor(
                module="dissect.target.plugins.apps.chat.msn",
                qualname="MSNPlugin",
                namespace="msn",
                path="apps.chat.msn",
                findable=True,
                functions=["history", "__call__"],
                exports=["history", "__call__"],
            ),
            # OS, regular
            "os.windows.generic.GenericPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.windows.generic",
                qualname="GenericPlugin",
                namespace=None,
                path="os.windows.generic",
                findable=True,
                functions=["domain"],
                exports=["domain"],
            ),
            "os.unix.cronjobs.CronjobPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix.cronjobs",
                qualname="CronjobPlugin",
                namespace=None,
                path="os.unix.cronjobs",
                findable=True,
                functions=["cronjobs"],
                exports=["cronjobs"],
            ),
            "os.windows.regf.runkeys.RunKeysPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.windows.regf.runkeys",
                qualname="RunKeysPlugin",
                namespace=None,
                path="os.windows.regf.runkeys",
                findable=True,
                functions=["runkeys"],
                exports=["runkeys"],
            ),
            # OS, shared name
            "os.unix.linux.services.ServicesPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix.linux.services",
                qualname="ServicesPlugin",
                namespace=None,
                path="os.unix.linux.services",
                findable=True,
                functions=["initd", "services", "systemd"],
                exports=["services"],
            ),
            "os.windows.services.ServicesPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.windows.services",
                qualname="ServicesPlugin",
                namespace=None,
                path="os.windows.services",
                findable=True,
                functions=["services"],
                exports=["services"],
            ),
            # OS, with default
            "os.default.locale.LocalePlugin": PluginDescriptor(
                module="dissect.target.plugins.os.default.locale",
                qualname="LocalePlugin",
                namespace=None,
                path="os.default.locale",
                findable=True,
                functions=["timezone"],
                exports=["timezone"],
            ),
            "os.unix.locale.UnixLocalePlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix.locale",
                qualname="UnixLocalePlugin",
                namespace=None,
                path="os.unix.locale",
                findable=True,
                functions=["timezone"],
                exports=["timezone"],
            ),
            "os.unix.linux.fortios.locale.FortiOSLocalePlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix.linux.fortios.locale",
                qualname="FortiOSLocalePlugin",
                namespace=None,
                path="os.unix.linux.fortios.locale",
                findable=True,
                functions=["timezone"],
                exports=["timezone"],
            ),
            "os.windows.locale.WindowsLocalePlugin": PluginDescriptor(
                module="dissect.target.plugins.os.windows.locale",
                qualname="WindowsLocalePlugin",
                namespace=None,
                path="os.windows.locale",
                findable=True,
                functions=["timezone"],
                exports=["timezone"],
            ),
        },
        __os__={
            "os.default._os.DefaultOSPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.default._os",
                qualname="DefaultOSPlugin",
                namespace=None,
                path="os.default._os",
                findable=True,
                functions=["hostname", "os", "os_tree"],
                exports=["hostname", "os"],
            ),
            "os.unix._os.UnixPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix._os",
                qualname="UnixPlugin",
                namespace=None,
                path="os.unix._os",
                findable=True,
                functions=["hostname", "domain", "os", "os_tree"],
                exports=["hostname", "domain", "os"],
            ),
            "os.unix.linux._os.LinuxPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix.linux._os",
                qualname="LinuxPlugin",
                namespace=None,
                path="os.unix.linux._os",
                findable=True,
                functions=["hostname", "domain", "os", "os_tree"],
                exports=["hostname", "domain", "os"],
            ),
            "os.unix.linux.fortios._os.FortiOSPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix.linux.fortios._os",
                qualname="FortiOSPlugin",
                namespace=None,
                path="os.unix.linux.fortios._os",
                findable=True,
                functions=["hostname", "domain", "os", "os_tree"],
                exports=["hostname", "domain", "os"],
            ),
            "os.windows._os.WindowsPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.windows._os",
                qualname="WindowsPlugin",
                namespace=None,
                path="os.windows._os",
                findable=True,
                functions=["hostname", "os", "os_tree"],
                exports=["hostname", "os"],
            ),
        },
        __child__={},
    ),
    __functions__=FunctionDescriptorLookup(
        __regular__={
            # Non-OS, regular
            "osinfo": {
                "general.osinfo.OSInfoPlugin": FunctionDescriptor(
                    name="osinfo",
                    namespace=None,
                    path="general.osinfo.osinfo",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="osinfo",
                    module="dissect.target.plugins.general.osinfo",
                    qualname="OSInfoPlugin",
                )
            },
            # Non-OS, with namespace
            "chat": {
                "apps.chat.chat.ChatPlugin": FunctionDescriptor(
                    name="chat",
                    namespace="chat",
                    path="apps.chat.chat",
                    exported=True,
                    internal=False,
                    findable=False,
                    alias=False,
                    output="record",
                    method_name="__call__",
                    module="dissect.target.plugins.apps.chat.chat",
                    qualname="ChatPlugin",
                )
            },
            "chat.history": {
                "apps.chat.chat.ChatPlugin": FunctionDescriptor(
                    name="chat.history",
                    namespace="chat",
                    path="apps.chat.chat.history",
                    exported=True,
                    internal=False,
                    findable=False,
                    alias=False,
                    output="record",
                    method_name="history",
                    module="dissect.target.plugins.apps.chat.chat",
                    qualname="ChatPlugin",
                )
            },
            "msn": {
                "apps.chat.msn.MSNPlugin": FunctionDescriptor(
                    name="msn",
                    namespace="msn",
                    path="apps.chat.msn",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="__call__",
                    module="dissect.target.plugins.apps.chat.msn",
                    qualname="MSNPlugin",
                )
            },
            "msn.history": {
                "apps.chat.msn.MSNPlugin": FunctionDescriptor(
                    name="msn.history",
                    namespace="msn",
                    path="apps.chat.msn.history",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="history",
                    module="dissect.target.plugins.apps.chat.msn",
                    qualname="MSNPlugin",
                )
            },
            # OS, regular
            "domain": {
                "os.windows.generic.GenericPlugin": FunctionDescriptor(
                    name="domain",
                    namespace=None,
                    path="os.windows.generic.domain",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="domain",
                    module="dissect.target.plugins.os.windows.generic",
                    qualname="GenericPlugin",
                )
            },
            "cronjobs": {
                "os.unix.cronjobs.CronjobPlugin": FunctionDescriptor(
                    name="cronjobs",
                    namespace=None,
                    path="os.unix.cronjobs.cronjobs",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="cronjobs",
                    module="dissect.target.plugins.os.unix.cronjobs",
                    qualname="CronjobPlugin",
                )
            },
            "runkeys": {
                "os.windows.regf.runkeys.RunKeysPlugin": FunctionDescriptor(
                    name="runkeys",
                    namespace=None,
                    path="os.windows.regf.runkeys.runkeys",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="runkeys",
                    module="dissect.target.plugins.os.windows.regf.runkeys",
                    qualname="RunKeysPlugin",
                )
            },
            # OS, shared name
            "services": {
                "os.windows.services.ServicesPlugin": FunctionDescriptor(
                    name="services",
                    namespace=None,
                    path="os.windows.services.services",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="services",
                    module="dissect.target.plugins.os.windows.services",
                    qualname="ServicesPlugin",
                ),
                "os.unix.linux.services.ServicesPlugin": FunctionDescriptor(
                    name="services",
                    namespace=None,
                    path="os.unix.linux.services.services",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="services",
                    module="dissect.target.plugins.os.unix.linux.services",
                    qualname="ServicesPlugin",
                ),
            },
            # OS, with default
            "timezone": {
                "os.default.locale.LocalePlugin": FunctionDescriptor(
                    name="timezone",
                    namespace=None,
                    path="os.default.locale.timezone",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="timezone",
                    module="dissect.target.plugins.os.default.locale",
                    qualname="LocalePlugin",
                ),
                "os.unix.locale.UnixLocalePlugin": FunctionDescriptor(
                    name="timezone",
                    namespace=None,
                    path="os.unix.locale.timezone",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="timezone",
                    module="dissect.target.plugins.os.unix.locale",
                    qualname="UnixLocalePlugin",
                ),
                "os.unix.linux.fortios.locale.FortiOSLocalePlugin": FunctionDescriptor(
                    name="timezone",
                    namespace=None,
                    path="os.unix.linux.fortios.locale.timezone",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="timezone",
                    module="dissect.target.plugins.os.unix.linux.fortios.locale",
                    qualname="FortiOSLocalePlugin",
                ),
                "os.windows.locale.WindowsLocalePlugin": FunctionDescriptor(
                    name="timezone",
                    namespace=None,
                    path="os.windows.locale.timezone",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="timezone",
                    module="dissect.target.plugins.os.windows.locale",
                    qualname="WindowsLocalePlugin",
                ),
            },
        },
        __os__={
            "hostname": {
                "os.default._os.DefaultOSPlugin": FunctionDescriptor(
                    name="hostname",
                    namespace=None,
                    path="os.default._os.hostname",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="hostname",
                    module="dissect.target.plugins.os.default._os",
                    qualname="DefaultOSPlugin",
                ),
                "os.unix._os.UnixPlugin": FunctionDescriptor(
                    name="hostname",
                    namespace=None,
                    path="os.unix._os.hostname",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="hostname",
                    module="dissect.target.plugins.os.unix._os",
                    qualname="UnixPlugin",
                ),
                "os.unix.linux._os.LinuxPlugin": FunctionDescriptor(
                    name="hostname",
                    namespace=None,
                    path="os.unix.linux._os.hostname",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="hostname",
                    module="dissect.target.plugins.os.unix.linux._os",
                    qualname="LinuxPlugin",
                ),
                "os.unix.linux.fortios._os.FortiOSPlugin": FunctionDescriptor(
                    name="hostname",
                    namespace=None,
                    path="os.unix.linux.fortios._os.hostname",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="hostname",
                    module="dissect.target.plugins.os.unix.linux.fortios._os",
                    qualname="FortiOSPlugin",
                ),
                "os.windows._os.WindowsPlugin": FunctionDescriptor(
                    name="hostname",
                    namespace=None,
                    path="os.windows._os.hostname",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="hostname",
                    module="dissect.target.plugins.os.windows._os",
                    qualname="WindowsPlugin",
                ),
            },
            "os": {
                "os.default._os.DefaultOSPlugin": FunctionDescriptor(
                    name="os",
                    namespace=None,
                    path="os.default._os.os",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="os",
                    module="dissect.target.plugins.os.default._os",
                    qualname="DefaultOSPlugin",
                ),
                "os.unix._os.UnixPlugin": FunctionDescriptor(
                    name="os",
                    namespace=None,
                    path="os.unix._os.os",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="os",
                    module="dissect.target.plugins.os.unix._os",
                    qualname="UnixPlugin",
                ),
                "os.unix.linux._os.LinuxPlugin": FunctionDescriptor(
                    name="os",
                    namespace=None,
                    path="os.unix.linux._os.os",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="os",
                    module="dissect.target.plugins.os.unix.linux._os",
                    qualname="LinuxPlugin",
                ),
                "os.unix.linux.fortios._os.FortiOSPlugin": FunctionDescriptor(
                    name="os",
                    namespace=None,
                    path="os.unix.linux.fortios._os.os",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="os",
                    module="dissect.target.plugins.os.unix.linux.fortios._os",
                    qualname="FortiOSPlugin",
                ),
                "os.windows._os.WindowsPlugin": FunctionDescriptor(
                    name="os",
                    namespace=None,
                    path="os.windows._os.os",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="os",
                    module="dissect.target.plugins.os.windows._os",
                    qualname="WindowsPlugin",
                ),
            },
            "domain": {
                "os.unix._os.UnixPlugin": FunctionDescriptor(
                    name="domain",
                    namespace=None,
                    path="os.unix._os.domain",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="domain",
                    module="dissect.target.plugins.os.unix._os",
                    qualname="UnixPlugin",
                ),
                "os.unix.linux._os.LinuxPlugin": FunctionDescriptor(
                    name="domain",
                    namespace=None,
                    path="os.unix.linux._os.domain",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="domain",
                    module="dissect.target.plugins.os.unix.linux._os",
                    qualname="LinuxPlugin",
                ),
                "os.unix.linux.fortios._os.FortiOSPlugin": FunctionDescriptor(
                    name="domain",
                    namespace=None,
                    path="os.unix.linux.fortios._os.domain",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="domain",
                    module="dissect.target.plugins.os.unix.linux.fortios._os",
                    qualname="FortiOSPlugin",
                ),
            },
        },
        __child__={},
    ),
    __ostree__={
        "os": {
            "default": {},
            "unix": {
                "linux": {
                    "fortios": {},
                }
            },
            "windows": {},
        }
    },
    __failed__=[],
)


@pytest.mark.parametrize(
    ("osfilter", "index", "expected_plugins"),
    [
        (
            None,
            "__regular__",
            [
                "dissect.target.plugins.general.osinfo",
                "dissect.target.plugins.apps.chat.chat",
                "dissect.target.plugins.apps.chat.msn",
                "dissect.target.plugins.os.windows.generic",
                "dissect.target.plugins.os.unix.cronjobs",
                "dissect.target.plugins.os.windows.regf.runkeys",
                "dissect.target.plugins.os.windows.services",
                "dissect.target.plugins.os.unix.linux.services",
                "dissect.target.plugins.os.default.locale",
                "dissect.target.plugins.os.unix.locale",
                "dissect.target.plugins.os.unix.linux.fortios.locale",
                "dissect.target.plugins.os.windows.locale",
            ],
        ),
        (
            None,
            "__os__",
            [
                "dissect.target.plugins.os.default._os",
                "dissect.target.plugins.os.unix._os",
                "dissect.target.plugins.os.unix.linux._os",
                "dissect.target.plugins.os.unix.linux.fortios._os",
                "dissect.target.plugins.os.windows._os",
            ],
        ),
        (
            "os.unix._os",
            "__regular__",
            [
                "dissect.target.plugins.general.osinfo",
                "dissect.target.plugins.apps.chat.chat",
                "dissect.target.plugins.apps.chat.msn",
                "dissect.target.plugins.os.unix.cronjobs",
                "dissect.target.plugins.os.unix.locale",
            ],
        ),
        (
            "os.unix.linux.fortios._os",
            "__regular__",
            [
                "dissect.target.plugins.general.osinfo",
                "dissect.target.plugins.apps.chat.chat",
                "dissect.target.plugins.apps.chat.msn",
                "dissect.target.plugins.os.unix.cronjobs",
                "dissect.target.plugins.os.unix.locale",
                "dissect.target.plugins.os.unix.linux.services",
                "dissect.target.plugins.os.unix.linux.fortios.locale",
            ],
        ),
        (
            "os.windows._os",
            "__regular__",
            [
                "dissect.target.plugins.general.osinfo",
                "dissect.target.plugins.apps.chat.chat",
                "dissect.target.plugins.apps.chat.msn",
                "dissect.target.plugins.os.windows.generic",
                "dissect.target.plugins.os.windows.regf.runkeys",
                "dissect.target.plugins.os.windows.services",
                "dissect.target.plugins.os.windows.locale",
            ],
        ),
        (
            "os.bar._os",
            "__regular__",
            [
                "dissect.target.plugins.general.osinfo",
                "dissect.target.plugins.apps.chat.chat",
                "dissect.target.plugins.apps.chat.msn",
            ],
        ),
        (
            "os.windows._os",
            "__os__",
            [],
        ),
    ],
)
def test_plugins(
    osfilter: str,
    index: str,
    expected_plugins: list[str],
) -> None:
    with (
        patch("dissect.target.plugin._get_plugins", return_value=MOCK_PLUGINS),
        patch("dissect.target.plugin._module_path", return_value=osfilter),
    ):
        if osfilter is not None:
            # osfilter must be a class or None
            osfilter = Mock

        plugin_descriptors = plugins(osfilter=osfilter, index=index)

        assert sorted(desc.module for desc in plugin_descriptors) == sorted(expected_plugins)


def test_plugins_default_plugin(target_default: Target) -> None:
    all_plugins = list(plugins())
    default_plugin_plugins = list(plugins(osfilter=target_default._os_plugin))

    assert default_plugin_plugins == all_plugins

    # The all_with_home is a sentinel function, which should be loaded for a
    # target with DefaultOSPlugin as OS plugin.
    sentinel_function = "all_with_home"
    has_sentinel_function = False
    for p in default_plugin_plugins:
        if sentinel_function in p.functions:
            has_sentinel_function = True
            break

    assert has_sentinel_function

    default_os_plugin_desc = plugins(osfilter=target_default._os_plugin, index="__os__")

    assert len(list(default_os_plugin_desc)) == 1


@pytest.mark.parametrize(
    ("os_plugin", "pattern", "expected_paths"),
    [
        # Direct match
        (DefaultOSPlugin, "osinfo", ["general.osinfo.osinfo"]),
        # Direct match with namespace, regardless of `findable` status
        (DefaultOSPlugin, "chat.history", ["apps.chat.chat.history"]),
        (DefaultOSPlugin, "msn.history", ["apps.chat.msn.history"]),
        # Find with tree search using wildcard
        (DefaultOSPlugin, "general.*", ["general.osinfo.osinfo"]),
        # Find with tree search using implicit wildcard
        (DefaultOSPlugin, "general.osinfo", ["general.osinfo.osinfo"]),
        # Find with exact tree match
        (DefaultOSPlugin, "general.osinfo.osinfo", ["general.osinfo.osinfo"]),
        # Find with tree search using more complicated patterns
        (DefaultOSPlugin, "general.[!o]*", []),
        (DefaultOSPlugin, "general.[!x]*", ["general.osinfo.osinfo"]),
        (DefaultOSPlugin, "general.??????.??????", ["general.osinfo.osinfo"]),
        # Namespaces do not match
        (DefaultOSPlugin, "chat.*", []),
        # Part of module paths do not match
        (DefaultOSPlugin, "generic", []),
        # OS direct match only matches within the same OS plugin
        (DefaultOSPlugin, "hostname", ["os.default._os.hostname"]),
        (WindowsPlugin, "hostname", ["os.windows._os.hostname"]),
        (LinuxPlugin, "hostname", ["os.unix.linux._os.hostname", "os.unix._os.hostname"]),
        # OS tree search only matches within the same OS plugin
        (DefaultOSPlugin, "os.windows._os.hostname", []),
        (WindowsPlugin, "os.default._os.hostname", []),
        (DefaultOSPlugin, "os.*.hostname", ["os.default._os.hostname"]),
        (WindowsPlugin, "os.*.hostname", ["os.windows._os.hostname"]),
        # # "os" hits the direct match, not the tree search
        (DefaultOSPlugin, "os", ["os.default._os.os"]),
        (WindowsPlugin, "os", ["os.windows._os.os"]),
        # Wildcard matches all regular functions and OS functions (within the same OS)
        (
            DefaultOSPlugin,
            "*",
            [
                "os.unix.linux.fortios.locale.timezone",
                "os.windows.regf.runkeys.runkeys",
                "os.unix.linux.services.services",
                "apps.chat.msn.history",
                "os.windows.generic.domain",
                "os.unix.cronjobs.cronjobs",
                "os.windows.services.services",
                "os.default.locale.timezone",
                "os.unix.locale.timezone",
                "os.windows.locale.timezone",
                "os.default._os.hostname",
                "os.default._os.os",
                "general.osinfo.osinfo",
            ],
        ),
        (
            WindowsPlugin,
            "*",
            [
                "os.windows.regf.runkeys.runkeys",
                "apps.chat.msn.history",
                "os.windows.generic.domain",
                "os.windows.services.services",
                "os.windows.locale.timezone",
                "os.windows._os.hostname",
                "os.windows._os.os",
                "general.osinfo.osinfo",
            ],
        ),
        # Some OS functions only exist on certain OS plugins
        (LinuxPlugin, "os.*.domain", ["os.unix.linux._os.domain", "os.unix._os.domain"]),
        # # Some plugins have complicated paths (i.e. LocalePlugin)
        (LinuxPlugin, "timezone", ["os.unix.locale.timezone"]),
        (FortiOSPlugin, "timezone", ["os.unix.linux.fortios.locale.timezone", "os.unix.locale.timezone"]),
        # Some plugins have overlapping names (i.e. "services")
        (FortiOSPlugin, "services", ["os.unix.linux.services.services"]),
        (WindowsPlugin, "services", ["os.windows.services.services"]),
    ],
)
def test_find_functions(os_plugin: type[OSPlugin], pattern: str, expected_paths: list[str]) -> None:
    with patch("dissect.target.plugin._get_plugins", return_value=MOCK_PLUGINS):
        target = Target()
        target._os_plugin = os_plugin

        found, _ = find_functions(pattern, target)
        assert [desc.path for desc in found] == expected_paths


def test_find_functions_windows(target_win: Target) -> None:
    found, _ = find_functions("services", target_win)

    assert len(found) == 1
    assert found[0].name == "services"
    assert found[0].path == "os.windows.services.services"


def test_find_functions_linux(target_linux: Target) -> None:
    found, _ = find_functions("services", target_linux)

    assert len(found) == 1
    assert found[0].name == "services"
    assert found[0].path == "os.unix.linux.services.services"


def test_find_functions_compatible_check(target_linux: Target) -> None:
    """Test if we correctly check for compatibility in ``find_functions`` and ``_filter_compatible``."""

    found, _ = find_functions("*", target_linux, compatibility=True)
    assert "os.unix.log.messages.syslog.syslog" not in [f"{f.path}.{f.name}" for f in found]

    with patch("dissect.target.plugins.apps.browser.chrome.ChromePlugin.check_compatible", return_value=None):
        found, _ = find_functions("*", target_linux, compatibility=True)
        functions = [f.path for f in found]
        assert "apps.browser.chrome.cookies" in functions
        assert "apps.browser.chrome.history" in functions


@pytest.mark.benchmark
def test_benchmark_functions_compatible_check(target_unix_users: Target, benchmark: BenchmarkFixture) -> None:
    """Benchmark ``_filter_compatible`` performance."""
    benchmark(lambda: find_functions("*", target_unix_users, compatibility=True))


def test_function_aliases(target_default: Target) -> None:
    """Test if alias functions are tagged as such correctly."""

    # function that is an alias should have an alias property set to True
    syslog_fd = find_functions("syslog", target_default)[0][0]
    assert syslog_fd
    assert syslog_fd.path == "os.unix.log.messages.syslog"
    assert syslog_fd.exported
    assert syslog_fd.alias

    # function that is not an alias should have an alias property set to False
    messages_fd = find_functions("messages", target_default)[0][0]
    assert messages_fd
    assert messages_fd.path == "os.unix.log.messages.messages"
    assert not messages_fd.alias


def test_function_required_arguments(target_default: Target) -> None:
    """Test if functions with required arguments are tagged as such correctly."""

    # function without any arguments should have an args property with an empty list
    syslog_fd = find_functions("syslog", target_default)[0][0]
    assert syslog_fd
    assert not syslog_fd.args

    # function with an argument should have an args property filled
    envfile_fd = find_functions("envfile", target_default)[0][0]
    assert envfile_fd
    assert envfile_fd.args == [
        (
            ("--env-path",),
            {
                "help": "path to scan environment files in",
                "required": True,
            },
        ),
        (
            ("--extension",),
            {
                "default": "env",
                "help": "extension of files to scan",
            },
        ),
    ]


def test_plugin_runtime_info() -> None:
    plugin_desc = next(p for p in plugins() if p.path == "general.users")
    assert plugin_desc.cls is UsersPlugin

    func_desc = next(p for p in functions() if p.path == "apps.other.env.envfile")
    assert func_desc.cls is EnvironmentFilePlugin
    assert func_desc.func is EnvironmentFilePlugin.envfile
    assert func_desc.record is EnvironmentFilePlugin.envfile.__record__
    assert func_desc.args == EnvironmentFilePlugin.envfile.__args__


def test_find_by_record_field_type(target_default: Target) -> None:
    assert "filesystem.walkfs.walkfs" in [desc.path for desc in find_functions_by_record_field_type("path")]
    assert "apps.other.env.envfile" in [
        desc.path for desc in find_functions_by_record_field_type("path", target_default, compatibility=True)
    ]

    with patch(
        "dissect.target.plugin.functions",
        return_value=[
            FunctionDescriptor(
                name="test",
                namespace=None,
                path="test",
                exported=True,
                internal=False,
                findable=True,
                alias=False,
                output="record",
                method_name="test",
                module="test",
                qualname="Test",
            )
        ],
    ):
        with pytest.raises(PluginError, match="An exception occurred while trying to load a plugin: test"):
            list(find_functions_by_record_field_type("path"))

        assert list(find_functions_by_record_field_type("path", ignore_load_errors=True)) == []


@pytest.mark.parametrize(
    "method_name",
    [
        "hostname",
        "ips",
        "version",
        "os",
        "architecture",
    ],
)
def test_os_plugin_property_methods(target_bare: Target, method_name: str) -> None:
    os_plugin = OSPlugin(target_bare)
    with pytest.raises(NotImplementedError):
        getattr(os_plugin, method_name)


class MockOS1(OSPlugin):
    __register__ = False

    @export(property=True)
    def hostname(self) -> str | None:
        pass

    @export(property=True)
    def ips(self) -> list[str]:
        pass

    @export(property=True)
    def version(self) -> str | None:
        pass

    @export(record=EmptyRecord)
    def users(self) -> list[Record]:
        pass

    @export(property=True)
    def os(self) -> str:
        pass

    @export(property=True)
    def architecture(self) -> str | None:
        pass


class MockOS2(OSPlugin):
    __register__ = False

    @export(property=True)
    def hostname(self) -> str | None:
        """Test docstring hostname."""

    @export(property=True)
    def ips(self) -> list[str]:
        """Test docstring ips."""

    @export(property=True)
    def version(self) -> str | None:
        """Test docstring version."""

    @export(record=EmptyRecord)
    def users(self) -> list[Record]:
        """Test docstring users."""

    @export(property=True)
    def os(self) -> str:
        """Test docstring os."""

    @export(property=True)
    def architecture(self) -> str | None:
        """Test docstring architecture."""


@pytest.mark.parametrize(
    ("subclass", "replaced"),
    [
        (MockOS1, True),
        (MockOS2, False),
    ],
)
def test_os_plugin___init_subclass__(subclass: type[OSPlugin], replaced: bool) -> None:
    exported_methods = {
        "hostname",
        "ips",
        "version",
        "users",
        "os",
        "architecture",
    }

    for method_name in exported_methods:
        os_method = getattr(OSPlugin, method_name)
        if isinstance(os_method, property):
            os_method = os_method.fget
        os_docstring = os_method.__doc__

        subclass_method = getattr(subclass, method_name)
        if isinstance(subclass_method, property):
            subclass_method = subclass_method.fget
        subclass_docstring = subclass_method.__doc__

        assert (os_docstring == subclass_docstring) is replaced
        if not replaced:
            assert subclass_docstring == f"Test docstring {method_name}."


@pytest.mark.parametrize(
    argnames=("os_plugin", "results"),
    argvalues=[
        (LinuxPlugin, ["linux", "unix"]),
        (FortiOSPlugin, ["fortios", "linux", "unix"]),
        (DebianPlugin, ["linux", "unix"]),
    ],
)
def test_os_tree(target_bare: Target, os_plugin: type[OSPlugin], results: list[str]) -> None:
    """Test if we correctly return the OS name tree."""
    target_bare._os_plugin = os_plugin
    target_bare.apply()
    assert target_bare.os_tree() == results


class _TestInternalPlugin(InternalPlugin):
    def test(self) -> None:
        pass


def test_internal_plugin() -> None:
    assert "test" not in _TestInternalPlugin.__exports__
    assert "test" in _TestInternalPlugin.__functions__


class _TestInternalNamespacePlugin(InternalNamespacePlugin):
    __namespace__ = "NS"

    def test(self) -> None:
        pass


def test_internal_namespace_plugin() -> None:
    assert "__subplugins__" in dir(_TestInternalNamespacePlugin)
    assert "test" not in _TestInternalNamespacePlugin.__exports__
    assert "test" in _TestInternalNamespacePlugin.__functions__


class ExampleFooPlugin(Plugin):
    """Example Foo Plugin."""

    __register__ = False

    def check_compatible(self) -> None:
        return

    @export
    @alias("bar")
    @alias(name="baz")
    def foo(self) -> Iterator[str]:
        """Yield foo!."""
        yield "foo!"


def test_plugin_alias(target_bare: Target) -> None:
    """Test ``@alias`` decorator behaviour."""
    target_bare.add_plugin(ExampleFooPlugin)
    assert target_bare.has_function("foo")
    assert target_bare.foo.__aliases__ == ["baz", "bar"]
    assert target_bare.has_function("bar")
    assert target_bare.has_function("baz")
    assert list(target_bare.foo()) == list(target_bare.bar()) == list(target_bare.baz())


@pytest.mark.parametrize(
    "descriptor",
    find_functions("*", Target(), compatibility=False, show_hidden=True)[0],
    ids=lambda d: d.path,
)
def test_exported_plugin_format(descriptor: FunctionDescriptor) -> None:
    """This test checks plugin style guide conformity for all exported plugins.

    References:
        - https://docs.dissect.tools/en/latest/contributing/style-guide.html
    """
    # Ignore DefaultOSPlugin, NamespacePlugin and OSPlugin instances
    if issubclass(descriptor.cls, (NamespacePlugin, OSPlugin)) or descriptor.cls is DefaultOSPlugin:
        return

    # Plugin method should specify what it returns
    assert descriptor.output in ["record", "yield", "default", "none"], (
        f"Invalid output_type for function {descriptor.func.__qualname__}"
    )

    annotations = None

    if hasattr(descriptor.func, "__annotations__"):
        annotations = descriptor.func.__annotations__

    elif isinstance(descriptor.func, property):
        annotations = descriptor.func.fget.__annotations__

    # Plugin method should have a return annotation
    assert annotations
    assert "return" in annotations, f"No return type annotation for function {descriptor.func.__qualname__}"

    # TODO: Check if the annotations make sense with the provided output_type

    # Plugin method should have a docstring
    method_doc_str = descriptor.func.__doc__
    assert isinstance(method_doc_str, str), f"No docstring for function {descriptor.func.__qualname__}"
    assert method_doc_str != "", f"Empty docstring for function {descriptor.func.__qualname__}"

    # The method docstring should compile to rst without warnings
    assert_valid_rst(method_doc_str)

    # The method docstring should follow our conventions
    assert_compliant_rst(method_doc_str)

    # Plugin class should have a docstring
    class_doc_str = descriptor.cls.__doc__
    assert isinstance(class_doc_str, str), f"No docstring for class {descriptor.cls.__name__}"
    assert class_doc_str != "", f"Empty docstring for class {descriptor.cls.__name__}"

    # The class docstring should compile to rst without warnings
    assert_valid_rst(class_doc_str)

    # The class docstring should follow our conventions
    assert_compliant_rst(class_doc_str)

    # Arguments of the plugin should define their type and if they are required (explicitly or implicitly).
    for arg in descriptor.args:
        names, settings = arg
        is_bool_action = settings.get("action", "") in (
            "store_true",
            "store_false",
        )

        assert names, f"No argument names for argument of function {descriptor.func.__qualname__}"
        assert sorted(names, key=len) == list(names), (
            f"Argument names {names!r} for function {descriptor.func.__qualname__} should specify short form first"
        )

        assert settings.get("default", 1) is not None, (
            f"Superfluous default of None for argument {names[0]} in function {descriptor.func.__qualname__}: "
            "default is implied as None already."
        )

        assert settings.get("help"), f"No help text for argument {names[0]} in function {descriptor.func.__qualname__}"

        dest = settings.get("dest") or names[-1].strip("-").replace("-", "_")
        assert dest in annotations, (
            f"Missing type annotation for argument {dest} in function {descriptor.func.__qualname__}"
        )

        # TODO: More strictly check type annotation, use a contains right now to also match optionals
        type_ = "bool" if is_bool_action else getattr(settings.get("type"), "__name__", "str")
        assert type_ in annotations[dest], (
            f"Invalid type annotation for argument {dest} in function {descriptor.func.__qualname__} "
            f"({annotations[dest]} instead of {type_})"
        )

        assert settings.get("type") is not str, (
            f"Superfluous type of str for argument {names[0]} in function {descriptor.func.__qualname__}: "
            "type is implied as str by default."
        )

        # Inverse checks

        if settings.get("required"):
            assert not settings.get("default"), (
                "It does not make sense to set an argument to required and have a default value"
                f"in {names[0]} in function {descriptor.func.__qualname__}"
            )

        if "required" in settings:
            assert settings.get("required"), (
                f"Superfluous required of False for argument {names[0]} in function {descriptor.func.__qualname__}: "
                "required is implied as False already."
            )

        if is_bool_action:
            assert "type" not in settings, (
                f"Type should not be set for store_true or store_false in {names[0]} in "
                f"function {descriptor.func.__qualname__}: "
                "type is implied as boolean already."
            )

            assert "default" not in settings, (
                f"Default should not be set for store_true or store_false in {names[0]} in "
                f"function {descriptor.func.__qualname__}: "
                "default is implied as opposite boolean already."
            )


def test_plugin_record_field_consistency() -> None:
    """Test if exported plugin functions yielding records do not have conflicting field names and types.

    For example, take the following TargetRecordDescriptors for plugin X, Y and Z::

        RecordX = TargetRecordDescriptor("record/x", [("varint", "my_field")])
        RecordY = TargetRecordDescriptor("record/y", [("path", "my_field")])
        RecordZ = TargetRecordDescriptor("record/y", [("string", "my_field")])

    The ``RecordX`` descriptor will fail in this test, since the field ``my_field`` cannot be of type ``varint``
    while also being used as ``string`` (and ``path``). The ``RecordY`` and ``RecordZ`` descriptors do not conflict,
    since the types ``path`` and ``string`` translate to the same ``wildcard`` type.

    Uses ``FIELD_TYPES_MAP`` which is loosely based on flow.record and ElasticSearch field types.

    References:
        - https://elastic.co/guide/en/elasticsearch/reference/current/mapping-types.html
        - https://github.com/fox-it/flow.record/tree/main/flow/record/fieldtypes
        - https://github.com/JSCU-NL/dissect-elastic
    """

    seen_field_names: set[str] = set()
    seen_field_types: dict[str, tuple[str | None, RecordDescriptor]] = {}
    inconsistencies: set[str] = set()

    FIELD_TYPES_MAP = {
        # strings
        "string": "string",
        "stringlist": "string",
        "wstring": "string",
        "path": "string",
        "uri": "string",
        "command": "string",
        "dynamic": "string",
        # ints
        "varint": "int",
        "filesize": "int",
        "uint32": "int",
        "uint16": "int",
        "float": "float",
        # ip / cidr
        "net.ipaddress": "ip",
        "net.ipnetwork": "ip_range",
        "net.ipinterface": "ip_range",
        # dates
        "datetime": "datetime",
        # other
        "boolean": "boolean",
        "bytes": "binary",
        "digest": "keyword",
    }

    for descriptor in find_functions("*", Target(), compatibility=False, show_hidden=True)[0]:
        # Test if plugin function record fields make sense and do not conflict with other records.
        if descriptor.output == "record" and hasattr(descriptor, "record"):
            # Functions can yield a single record or a list of records.
            records = descriptor.record if isinstance(descriptor.record, list) else [descriptor.record]

            for record in records:
                assert isinstance(record, RecordDescriptor), (
                    f"{record!r} of function {descriptor!r} is not of type RecordDescriptor"
                )
                if record.name != "empty":
                    assert record.fields, f"{record!r} has no fields"

                for name, field in record.fields.items():
                    # Make sure field names have the same type when translated. This check does not save multiple field
                    # name and typenames, this is a bare-minumum check only.

                    # We only care about the field type, not if it is a list of that type.
                    field_typename = field.typename.replace("[]", "")

                    assert field_typename in FIELD_TYPES_MAP, (
                        f"Field type {field_typename} is not mapped in FIELD_TYPES_MAP, please add it manually."
                    )

                    if name in seen_field_names:
                        seen_typename, seen_record = seen_field_types[name]
                        if FIELD_TYPES_MAP[seen_typename] != FIELD_TYPES_MAP[field_typename]:
                            inconsistencies.add(
                                f"<{record.name} ({field.typename!r}, '{name}')> is duplicate mismatch of <{seen_record.name} ({seen_typename!r}, '{name}')>"  # noqa: E501
                            )

                    else:
                        seen_field_names.add(name)
                        seen_field_types[name] = (field_typename, record)

    if inconsistencies:
        pytest.fail(
            f"Found {len(inconsistencies)} inconsistencies in RecordDescriptors:\n" + "\n".join(inconsistencies)
        )


def assert_valid_rst(src: str) -> None:
    """Attempts to compile the given string to rst."""

    try:
        publish_string(src, settings_overrides={"halt_level": 2})

    except SystemMessage as e:
        # Limited context was provided to docutils, so some exceptions could incorrectly be raised.
        # We can assume that if the rst is truly invalid this will also be caught by `tox -e build-docs`.
        if "Unknown interpreted text role" not in str(e):
            pytest.fail(f"Invalid rst: {e}", pytrace=False)


def assert_compliant_rst(src: str) -> None:
    """Makes sure that the given rst docstring follows the project's conventions."""

    # Explicit message stating we want References instead of Resources to prevent confusion
    if "Resources:\n" in src:
        pytest.fail(f"Invalid rst: docstring contains 'Resources' instead of 'References': {src!r}", pytrace=False)

    # Generic message stating lists should start with References (assumes lists always have at least one 'http')
    if "- http" in src and "References:\n" not in src:
        pytest.fail(f"Invalid rst: docstring contains list but does not mention 'References': {src!r}", pytrace=False)

    # Make sure we use stripes instead of bullets (assumes lists always have at least one 'http')
    if "* http" in src:
        pytest.fail(f"Invalid rst: docstring contains bullet instead of dash in list: {src!r}", pytrace=False)


@pytest.mark.parametrize(
    "descriptor",
    [descriptor for descriptor in plugins() if descriptor.namespace and "." in descriptor.namespace],
    ids=lambda descriptor: descriptor.namespace,
)
def test_nested_namespace_consistency(descriptor: PluginDescriptor) -> None:
    """Test whether all parts of nested namespaces exist and that there are no conflicts with other functions."""

    parts = descriptor.namespace.split(".")
    for i in range(len(parts)):
        part = ".".join(parts[: i + 1])
        result = list(lookup(part))

        if not result:
            pytest.fail(f"Unreachable namespace {descriptor.namespace!r}, namespace {part!r} does not exist.")

        if len(result) > 1:
            conflicts = ", ".join(
                f"{desc.name} ({desc.module}.{desc.qualname})" for desc in result if desc.namespace != part
            )
            pytest.fail(f"Namespace name {descriptor.namespace!r} has conflicts with function name: {conflicts}")


@pytest.mark.parametrize(
    "descriptor",
    # Match plugin classes which are a *direct* base of NamespacePlugin only using :meth:`Plugin.__bases__`,
    # instead of using ``issubclass`` which would also yield indirectly inherited Plugin classes.
    [descriptor for descriptor in plugins() if NamespacePlugin in descriptor.cls.__bases__],
    ids=lambda descriptor: descriptor.qualname,
)
def test_namespace_class_usage(descriptor: PluginDescriptor) -> None:
    """This test checks if :class:`NamespacePlugin` usage is correct.

    :class:`NamespacePlugin` is reserved for "grouping" other plugins of the same category. See for
    example :class:`BrowserPlugin` or :class:`WebserverPlugin`.

    If you want to expose plugin functions under a shared name, e.g. ``foo.bar`` and ``foo.baz``,
    you should use :class:`Plugin` with ``Plugin.__namespace__ = "foo"`` instead.

    Resources:
        - https://github.com/fox-it/dissect.target/issues/1180
    """

    assert descriptor.cls.__subclasses__(), (
        f"NamespacePlugin {descriptor.module}.{descriptor.qualname} has no subclasses, are you sure you're using NamespacePlugin correctly?"  # noqa: E501
    )



from __future__ import annotations

from unittest.mock import Mock, patch

import pytest

from dissect.target.loaders.phobos import (
    EXTFS_NEEDLE,
    EXTFS_NEEDLE_OFFSET,
    NTFS_NEEDLE,
    PhobosLoader,
)


@pytest.mark.parametrize(
    ("needle", "offset", "expected_offset", "element_count", "element_size"),
    [
        (
            EXTFS_NEEDLE,
            EXTFS_NEEDLE_OFFSET + 1337,
            1337,
            1024,
            1024,
        ),
        (
            NTFS_NEEDLE,
            1337,
            1337,
            1024,
            1024,
        ),
    ],
)
def test_phobos_loader_map(
    needle: bytes,
    offset: int,
    expected_offset: int,
    element_count: int,
    element_size: int,
) -> None:
    mock_target = Mock()
    phobos_loader = PhobosLoader(mock_target)
    mock_fh = Mock()
    mock_fs = Mock()
    mock_fs.ntfs.boot_sector.NumberSectors = element_count
    mock_fs.ntfs.sector_size = element_size
    mock_fs.extfs.block_count = element_count
    mock_fs.extfs.block_size = element_size
    expected_size = element_count * element_size

    with (
        patch.object(phobos_loader.path, "open", return_value=mock_fh),
        patch(
            "dissect.target.loaders.phobos.scrape_pos",
            return_value=[(needle, offset)],
            autospec=True,
        ),
        patch("dissect.util.stream.RelativeStream", autospec=True) as mock_stream,
        patch(
            "dissect.target.filesystem.open",
            return_value=mock_fs,
            autospec=True,
        ),
    ):
        phobos_loader.map(mock_target)
        mock_stream.assert_called_with(mock_fh, expected_offset)
        mock_target.filesystems.add.assert_called_with(mock_fs)
        mock_target.fs.mount.assert_called_with("fs0", mock_fs)
        mock_fh.seek.assert_called_with(offset + expected_size)



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest
from dissect.thumbcache.exceptions import UnknownThumbnailTypeError

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugins.os.windows.thumbcache import ThumbcachePlugin

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


def create_user_data_paths(target_win: Target, tmp_path: Path) -> Path:
    """Add the explorer userdata path to ``target_win``.

    Returns: the path of created directory.
    """

    mocked_user_details = Mock()
    mocked_user = Mock()
    mocked_user.home_path = tmp_path
    mocked_user_details.all_with_home.return_value = [mocked_user]
    target_win.user_details = mocked_user_details

    explorer_dir = tmp_path / "appdata/local/microsoft/windows/explorer/"
    explorer_dir.mkdir(parents=True)

    return explorer_dir


def test_thumbcache_unsupported(target_win: Target, tmp_path: Path) -> None:
    with pytest.raises(UnsupportedPluginError):
        target_win.add_plugin(ThumbcachePlugin)

    create_user_data_paths(target_win, tmp_path)

    # Raises an error as there are no *_idx files.
    with pytest.raises(UnsupportedPluginError):
        target_win.add_plugin(ThumbcachePlugin)


def test_thumbcach_supported(target_win: Target, tmp_path: Path) -> None:
    explorer_dir = create_user_data_paths(target_win, tmp_path)

    (explorer_dir / "thumbcache_idx.db").touch()
    target_win.add_plugin(ThumbcachePlugin)


def test_thumbcache_dump_entry(target_win: Target, tmp_path: Path) -> None:
    create_user_data_paths(target_win, tmp_path)
    plugin = ThumbcachePlugin(target_win)
    with patch("dissect.target.plugins.os.windows.thumbcache.dump_entry_data_through_index") as mocked_dump:
        list(plugin._parse_thumbcache(None, None, Mock()))
        mocked_dump.assert_called_once()


def test_thumbcache_create_flow(target_win: Target, tmp_path: Path) -> None:
    create_user_data_paths(target_win, tmp_path)
    with patch("dissect.target.plugins.os.windows.thumbcache.Thumbcache") as mocked_thumbcache:
        plugin = ThumbcachePlugin(target_win)
        plugin._create_entries = Mock(return_value=["Hello"])
        assert len(list(plugin._parse_thumbcache(None, None, None))) == 1
        mocked_thumbcache.assert_called_once()


def test_thumbcache_unknown_exception(target_win: Target, tmp_path: Path) -> None:
    create_user_data_paths(target_win, tmp_path)
    plugin = ThumbcachePlugin(target_win)
    logger = Mock()
    target_win.log = logger

    with patch("dissect.target.plugins.os.windows.thumbcache.Thumbcache"):
        plugin._create_entries = Mock(side_effect=[FileNotFoundError])
        list(plugin._parse_thumbcache(None, None, None))
        logger.error.assert_called_once()
        logger.debug.assert_called_once()


def test_thumbcache_thumbcache_exception(target_win: Target, tmp_path: Path) -> None:
    create_user_data_paths(target_win, tmp_path)
    plugin = ThumbcachePlugin(target_win)
    logger = Mock()
    target_win.log = logger

    with patch("dissect.target.plugins.os.windows.thumbcache.Thumbcache"):
        plugin._create_entries = Mock(side_effect=[UnknownThumbnailTypeError])
        list(plugin._parse_thumbcache(None, None, None))
        logger.error.assert_called_once()



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.os.unix.linux.debian.apt import AptPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.mark.parametrize(
    "test_file",
    [
        "history.log",
        "history.log.1.gz",
        "history.log.1.bz2",
    ],
)
def test_apt_logs(test_file: str, target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    tz = timezone.utc
    data_file = absolute_path(f"_data/plugins/os/unix/linux/debian/apt/{test_file}")
    fs_unix.map_file(f"/var/log/apt/{test_file}", data_file)
    target_unix.add_plugin(AptPlugin)

    results = list(target_unix.apt.logs())
    assert len(results) == 18

    for record in results:
        assert record.package_manager == "apt"

    assert results[0].ts == datetime(2022, 9, 2, 6, 36, 31, tzinfo=tz)
    assert results[0].operation == "update"
    assert results[0].package_name == "libcurl4:amd64 (7.68.0-1ubuntu2.12, 7.68.0-1ubuntu2.13)"
    assert results[0].command == "/usr/bin/unattended-upgrade"
    assert results[0].requested_by_user is None

    assert results[-1].ts == datetime(2022, 9, 7, 7, 48, 28, tzinfo=tz)
    assert results[-1].operation == "update"
    assert results[-1].package_name == "linux-generic:amd64 (5.4.0.125.126, 5.4.0.126.127)"
    assert results[-1].command == "/usr/bin/unattended-upgrade"
    assert results[-1].requested_by_user == "user (1000)"



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

from dissect.cstruct import cstruct

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, arg, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target

bitmap_cache_def = """
// https://www.cert.ssi.gouv.fr/actualite/CERTFR-2016-ACT-017/
struct bin_header {
    CHAR    magic[8];
    DWORD   version;
};

struct bin_tile_header {
    DWORD   key1;
    DWORD   key2;
    WORD    tile_width;
    WORD    tile_height;
};

struct bmc_tile_header {
    DWORD   key1;
    DWORD   key2;
    WORD    tile_width;
    WORD    tile_height;
    DWORD   tile_length;
    DWORD   tile_params_unk_1: 3;
    DWORD   tile_params_compression: 1;
    DWORD   tile_params_unk_2: 28;
};
"""

bitmap_def = """
// https://stackoverflow.com/questions/20864752/how-is-defined-the-data-type-fxpt2dot30-in-the-bmp-file-structure
typedef LONG FXPT2DOT30;

// https://learn.microsoft.com/en-us/windows/win32/api/wingdi/ns-wingdi-ciexyz
struct CIEXYZ {
    FXPT2DOT30  ciexyzX;
    FXPT2DOT30  ciexyzY;
    FXPT2DOT30  ciexyzZ;
};

// https://learn.microsoft.com/en-us/windows/win32/api/wingdi/ns-wingdi-ciexyztriple
struct CIEXYZTRIPLE {
    CIEXYZ  ciexyzRed;
    CIEXYZ  ciexyzGreen;
    CIEXYZ  ciexyzBlue;
};

// https://learn.microsoft.com/en-us/windows/win32/api/wingdi/ns-wingdi-bitmapfileheader
struct BITMAPFILEHEADER {
    WORD    bfType;
    DWORD   bfSize;
    WORD    bfReserved1;
    WORD    bfReserved2;
    DWORD   bfOffBits;
};

// https://learn.microsoft.com/en-us/windows/win32/api/wingdi/ns-wingdi-bitmapv4header
struct BITMAPV4HEADER {
    DWORD           bV4Size;
    LONG            bV4Width;
    LONG            bV4Height;
    WORD            bV4Planes;
    WORD            bV4BitCount;
    DWORD           bV4V4Compression;
    DWORD           bV4SizeImage;
    LONG            bV4XPelsPerMeter;
    LONG            bV4YPelsPerMeter;
    DWORD           bV4ClrUsed;
    DWORD           bV4ClrImportant;
    DWORD           bV4RedMask;
    DWORD           bV4GreenMask;
    DWORD           bV4BlueMask;
    DWORD           bV4AlphaMask;
    DWORD           bV4CSType;
    CIEXYZTRIPLE    bV4Endpoints;
    DWORD           bV4GammaRed;
    DWORD           bV4GammaGreen;
    DWORD           bV4GammaBlue;
};
"""

c_bitmap_cache = cstruct().load(bitmap_cache_def)
c_bmp = cstruct().load(bitmap_def)

BMP_MAGIC = int.from_bytes(b"BM", "little")
BIN_MAGIC = b"RDP8bmp\x00"

BMP_DATA_OFFSET = len(c_bmp.BITMAPFILEHEADER) + len(c_bmp.BITMAPV4HEADER)  # 122

# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-wmf/eb4bbd50-b3ce-4917-895c-be31f214797f
LCS_WINDOWS_COLOR_SPACE = int.from_bytes(b"Win ", "big")

# This member is ignored unless the bV4CSType member specifies LCS_CALIBRATED_RGB. We therefore prepare an empty logical
# color space for when we export our BMPs
EMPTY_LOGICAL_COLOR_SPACE = c_bmp.CIEXYZTRIPLE(
    ciexyzRed=c_bmp.CIEXYZ(ciexyzX=0, ciexyzY=0, ciexyzZ=0),
    ciexyzGreen=c_bmp.CIEXYZ(ciexyzX=0, ciexyzY=0, ciexyzZ=0),
    ciexyzBlue=c_bmp.CIEXYZ(ciexyzX=0, ciexyzY=0, ciexyzZ=0),
)
EMPTY_PIXEL = b"\xff\xff\xff\x00"  # Transparent white pixel
BORDER_PIXEL = b"\x80\x80\x80\xff"  # Grey pixel


RDPCacheRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/rdp/cache",
    [
        ("datetime", "ctime"),
        ("datetime", "mtime"),
        ("path", "path"),
    ],
)


@dataclass
class BitmapTile:
    width: int
    height: int
    colors: bytes
    is_remnant: bool = False


def parse_color_data(data: bytes, reverse_rows: bool = False, row_width: int = 64) -> bytes:
    """Parse bitmap color data.

    Optionally can reverse the row order of the bitmap data, which is useful when parsing a
    bitmap that is top-down when you want it to be bottom-up (like in .bin files). Assumes 32 bits-per-pixel.
    """
    color_data = b""
    row = b""
    while len(data) > 0:
        chunk = data[:3] + b"\xff"
        data = data[4:]
        if reverse_rows:
            row += chunk
            if len(row) % (row_width * 4) == 0:
                # We have read a row
                # Prepend the row to the output data, essentially reversing the row order
                color_data = row + color_data
                row = b""
        else:
            color_data += chunk
    return color_data


def tile_to_bitmap(tile: BitmapTile) -> bytes:
    """Given a tile, convert it to a valid bitmap file."""
    file_header_new = c_bmp.BITMAPFILEHEADER(
        bfType=BMP_MAGIC,
        bfSize=len(tile.colors) + BMP_DATA_OFFSET,
        bfReserved1=0,
        bfReserved2=0,
        bfOffBits=BMP_DATA_OFFSET,
    )

    info_header_new = c_bmp.BITMAPV4HEADER(
        bV4Size=len(c_bmp.BITMAPV4HEADER),
        bV4Width=tile.width,
        bV4Height=tile.height,
        bV4Planes=1,
        bV4BitCount=32,
        bV4V4Compression=3,
        bV4SizeImage=len(tile.colors),
        bV4XPelsPerMeter=0,
        bV4YPelsPerMeter=0,
        bV4ClrUsed=0,
        bV4ClrImportant=0,
        bV4RedMask=0xFF0000,
        bV4GreenMask=0xFF00,
        bV4BlueMask=0xFF,
        bV4AlphaMask=0xFF000000,
        bV4CSType=LCS_WINDOWS_COLOR_SPACE,
        bV4Endpoints=EMPTY_LOGICAL_COLOR_SPACE,
        bV4GammaRed=0,
        bV4GammaGreen=0,
        bV4GammaBlue=0,
    )
    return file_header_new.dumps() + info_header_new.dumps() + tile.colors


def wrap_square_colors_in_border(colors: bytes, side_length: int, border_pixel: bytes, border_thickness: int) -> bytes:
    """Wrap color data in a colored-border."""
    while border_thickness > 0:
        border_thickness -= 1

        # Add a bottom row, Add a top row, Add a left-row and a right-row
        border_row = border_pixel * (side_length + 2)
        new_colors = b""
        for i in range(side_length):
            offset = i * side_length * 4  # 4 bytes per pixel

            color_slice = colors[offset : offset + side_length * 4]

            # Add a border-pixel to the left and right of this 'stripe'
            new_colors += border_pixel + color_slice + border_pixel

        # Add a border-stripe on top and on the bottom
        colors = border_row + new_colors + border_row

        # We have increased the size of the tile with 2 (one pixel on each end)
        side_length += 2

    return colors


def assemble_tiles_into_collage(tiles: list[BitmapTile], border_around_tile: int = 0) -> BitmapTile:
    """Assemble a list of tiles into one tile containing all color data."""
    tiles_in_row = 64
    rows: list[list[bytes]] = [[]]
    current_row = 0

    if not isinstance(border_around_tile, int) or border_around_tile < 0:
        raise ValueError("Argument border_around_tile should be zero or a positive integer.")

    tile_length = 64 + (border_around_tile * 2)

    padding_tile = EMPTY_PIXEL * tile_length * tile_length

    for tile in tiles:
        padding_color_data = ((64 * 64) - (tile.width * tile.height)) * EMPTY_PIXEL
        tile_color_data = tile.colors + padding_color_data

        if len(rows[current_row]) == tiles_in_row:
            current_row += 1
            rows.append([])

        if border_around_tile > 0:
            tile_color_data = wrap_square_colors_in_border(tile_color_data, 64, BORDER_PIXEL, border_around_tile)
        rows[current_row].append(tile_color_data)

    # If we only have one row (because there are not enough tiles) we decrease the collage width
    tiles_in_row = min(tiles_in_row, len(tiles))

    # Add padding tiles to the last row
    while len(rows[current_row]) < tiles_in_row:
        rows[current_row].append(padding_tile)

    final_image: list[bytes] = []
    for row in rows:
        # We can't dump every tile's color data one after another, as bitmaps go row-by-row. We 'slice' each horizontal
        # line from each tile, concat them to one another into one line of the final collage, and add that line to the
        # final image
        for i in range(tile_length):
            for tile_color_data in row:
                tile_data_size = tile_length * 4  # 4 bytes per pixel
                offset = i * tile_data_size
                line_from_tile = tile_color_data[offset : offset + tile_data_size]
                final_image.append(line_from_tile)

    collage_width = tiles_in_row * tile_length
    collage_height = tile_length * len(rows)
    return BitmapTile(collage_width, collage_height, b"".join(final_image))


def extract_bin(fh: BinaryIO) -> Iterator[BitmapTile]:
    """Extract bitmap tiles from a Cache000[1-4].bin bitmap cache file.

    These files are found on modern Windows versions.
    """
    fh.seek(0)
    bin_header = c_bitmap_cache.bin_header(fh)
    if bin_header.magic != BIN_MAGIC:
        raise ValueError(f"Invalid BIN file magic: {bin_header.magic}")

    while True:
        try:
            tile_header = c_bitmap_cache.bin_tile_header(fh)
        except EOFError:
            break
        tile_length = 4 * tile_header.tile_width * tile_header.tile_height
        colors = parse_color_data(fh.read(tile_length), reverse_rows=True, row_width=tile_header.tile_width)
        yield BitmapTile(tile_header.tile_width, tile_header.tile_height, colors)


def extract_bmc(fh: BinaryIO) -> Iterator[BitmapTile]:
    """Extract bitmap cache from bmc files, which are typically found on older Windows versions."""
    fh.seek(0)
    while True:
        try:
            tile_header = c_bitmap_cache.bmc_tile_header(fh)
        except EOFError:
            break

        if tile_header.tile_params_compression:
            raise NotImplementedError("No support for compressed bmc files")

        bytes_per_pixel = tile_header.tile_length // (tile_header.tile_width * tile_header.tile_height)
        if bytes_per_pixel != 4:
            if not (0 <= bytes_per_pixel <= 4):
                raise ValueError(f"Unexpected bpp value {bytes_per_pixel}")
            else:
                raise NotImplementedError("No support for bmc files with a bits-per-pixel other than 32")

        tile_size = bytes_per_pixel * tile_header.tile_width * tile_header.tile_height
        tile_colors = parse_color_data(fh.read(tile_size))
        yield BitmapTile(tile_header.tile_width, tile_header.tile_height, tile_colors)

        # Unlike bin-files, bmc files are pre-allocated with space for 64x64 pixel tiles. Whenever a tile is not exactly
        # 64 by 64, the remaining bytes until the next allocated slot are likely to contain leftover data from (a)
        # previous tile(s). bmc-tools calls these artefacts 'old' tiles, we call them 'remnant' tiles

        old_color_data_length = (64 * 64 - (tile_header.tile_width * tile_header.tile_height)) * bytes_per_pixel
        if old_color_data_length != 0:
            remnant_tile_colors = fh.read(old_color_data_length)
            if remnant_tile_colors == b"\x00" * old_color_data_length:
                # It's only null bytes, there is no content here
                continue
            if remnant_tile_colors == b"":
                # We have reached the end of the file, there's no color data after this tile
                break
            yield BitmapTile(
                64 - tile_header.tile_width,
                64 - tile_header.tile_height,
                remnant_tile_colors,
                is_remnant=True,
            )


class RdpCachePlugin(Plugin):
    """Plugin to extract the RDP Bitmap Cache from a Windows target.

    References:
        - https://www.cert.ssi.gouv.fr/actualite/CERTFR-2016-ACT-017/
    """

    __namespace__ = "rdpcache"

    CACHE_PATH = "AppData/Local/Microsoft/Terminal Server Client/Cache/"
    GLOBS = ("Cache*.bin", "bcache2*.bmc")

    def __init__(self, target: Target):
        super().__init__(target)
        self._caches: list[tuple[UserDetails, TargetPath]] = []

        for user_details in self.target.user_details.all_with_home():
            user_rdp_cache = user_details.home_path.joinpath(self.CACHE_PATH)
            for glob in self.GLOBS:
                for cache_path in user_rdp_cache.glob(glob):
                    if cache_path.lstat().st_size > 0:
                        self._caches.append((user_details, cache_path))

    def check_compatible(self) -> None:
        """At least one bitmap cache file with contents is necessary."""
        if not self._caches:
            raise UnsupportedPluginError("No RDP Cache files found on target")

    def _recover_tiles_from_cache(self, path: TargetPath) -> Iterator[BitmapTile]:
        """Given a target path, select the correct extraction method and yield bitmap tiles."""
        fh = path.open("rb")
        if fh.read(len(BIN_MAGIC)) == BIN_MAGIC:
            yield from extract_bin(fh)
        else:
            yield from extract_bmc(fh)

    @export(record=RDPCacheRecord)
    def paths(self) -> Iterator[RDPCacheRecord]:
        """Yield paths and timestamps of RDP Cache bitmap files."""
        for user_details, path in self._caches:
            lstat = path.lstat()
            yield RDPCacheRecord(
                ctime=lstat.st_ctime,
                mtime=lstat.st_mtime,
                path=path,
                _user=user_details.user,
                _target=self.target,
            )

    @arg("-o", "--output", dest="output_dir", type=Path, required=True, help="path to recover bitmap files to")
    @arg(
        "-g",
        "--grid",
        dest="as_grid",
        action="store_true",
        help="assemble all recovered tiles of a cache into a single grid image with borders around each tile",
    )
    @arg(
        "-c",
        "--collage",
        dest="as_collage",
        action="store_true",
        help="assemble all recovered tiles of a cache into one image",
    )
    @arg(
        "-n",
        "--no-tiles",
        dest="no_individual_tiles",
        action="store_true",
        help="do not save each recovered tile as a separate bitmap",
    )
    @arg(
        "-r",
        "--remnants",
        choices=["only", "include", "exclude"],
        default="exclude",
        help="include old leftover colordata in between tiles as separate, 'remnant' tiles",
    )
    @export(output="none")
    def recover(
        self, output_dir: Path, no_individual_tiles: bool, as_collage: bool, as_grid: bool, remnants: str
    ) -> None:
        """Extract bitmaps from Windows' RDP Client cache files."""

        if not output_dir.exists():
            self.target.log.error("Destination folder does not exist: %s", output_dir)
            return

        if not as_grid and not as_collage and no_individual_tiles:
            self.target.log.error("No export setting specified (one of --grid, --collage, --no-tiles)")
            return

        include_remnants = remnants == "only" or remnants == "include"
        exclude_regular_tiles = remnants == "only"

        for user_details, path in self._caches:
            prefix = f"{self.target.name}_{user_details.user.name}_{path.name}_"
            regular_tiles = []
            remnant_tiles = []
            for i, tile in enumerate(self._recover_tiles_from_cache(path)):
                if tile.is_remnant:
                    remnant_tiles.append(tile)
                    if not include_remnants:
                        continue
                    filename = f"{prefix}remnant_{i}.bmp"
                else:
                    regular_tiles.append(tile)
                    if exclude_regular_tiles:
                        continue
                    filename = f"{prefix}{i}.bmp"

                if not no_individual_tiles:
                    output_dir.joinpath(filename).write_bytes(tile_to_bitmap(tile))

            if include_remnants and not remnant_tiles:
                self.target.log.info("No remnant tile data found for %s", path)

            elif remnant_tiles and not include_remnants:
                self.target.log.warning("Remnant tile data found for %s, but remnant setting is set to exclude", path)

            if path.name.endswith(".bin"):
                # For bin files, collages look better if we assemble them in reverse order
                regular_tiles.reverse()
                remnant_tiles.reverse()

            assemble_settings = []
            if as_collage:
                assemble_settings.append(("collage", 0))
            if as_grid:
                assemble_settings.append(("grid", 2))

            valid_tile_sets = []
            if regular_tiles and not exclude_regular_tiles:
                valid_tile_sets.append(("", regular_tiles))

            if remnant_tiles and include_remnants:
                valid_tile_sets.append(("remnants_", remnant_tiles))

            for assemble_name, border_around_tile in assemble_settings:
                for tile_type_name, tiles in valid_tile_sets:
                    file = output_dir.joinpath(f"{prefix}{tile_type_name}{assemble_name}.bmp")
                    try:
                        file.write_bytes(tile_to_bitmap(assemble_tiles_into_collage(tiles, border_around_tile)))
                    except Exception as e:
                        self.target.log.error("Unable to write bitmap to file %s: %s", file, e)  # noqa: TRY400
                        self.target.log.debug("", exc_info=e)



from __future__ import annotations

import pytest
from dissect.regf import c_regf

from dissect.target.helpers.regutil import (
    HiveCollection,
    KeyCollection,
    RegFlex,
    RegistryHive,
    RegistryKeyNotFoundError,
    RegistryValueType,
    VirtualHive,
    VirtualKey,
    glob_ext,
    glob_ext0,
    glob_ext1,
    glob_split,
    has_glob_magic,
)
from tests._utils import absolute_path


def test_regflex() -> None:
    regflex = RegFlex()

    with absolute_path("_data/helpers/regutil/regflex.reg").open() as fh:
        regflex.map_definition(fh)

    assert "HKEY_CURRENT_USER" in regflex.hives
    hive = regflex.hives["HKEY_CURRENT_USER"]
    assert hive.key("Test1").value("Value1").value == "a"

    assert "HKEY_CLASSES_ROOT" in regflex.hives
    hive = regflex.hives["HKEY_CLASSES_ROOT"]
    assert hive.key("Test2").value("Value2").value == "b"

    assert "HKEY_LOCAL_MACHINE" in regflex.hives
    hive = regflex.hives["HKEY_LOCAL_MACHINE"]
    key = hive.key("Test Values")
    assert len(key.values()) == 15
    assert key.value("String").value == "a"

    assert key.value("String").type == RegistryValueType.SZ
    assert key.value("Long String").value == "a" * 1218
    assert key.value("Long String").type == RegistryValueType.SZ

    assert key.value("Binary").value == b"\x00" * 8
    assert key.value("Binary").type == RegistryValueType.BINARY

    assert key.value("Long Binary").value == b"\x00" * 240
    assert key.value("Long Binary").type == RegistryValueType.BINARY

    assert key.value("Dword").value == 1
    assert key.value("Dword").type == RegistryValueType.DWORD

    assert key.value("None").value == b"\x00\x01\x02\x04"
    assert key.value("None").type == RegistryValueType.NONE

    assert key.value("Hex String").value == "abcd"
    assert key.value("Hex String").type == RegistryValueType.SZ

    assert key.value("Expandable String").value == "a"
    assert key.value("Expandable String").type == RegistryValueType.EXPAND_SZ

    assert key.value("Long Expandable String").value == "a" * 2584
    assert key.value("Long Expandable String").type == RegistryValueType.EXPAND_SZ

    assert key.value("Hex Binary").value == b"\x00" * 8
    assert key.value("Hex Binary").type == RegistryValueType.BINARY

    assert key.value("Hex Dword LE").value == 1
    assert key.value("Hex Dword LE").type == RegistryValueType.DWORD

    assert key.value("Hex Dword BE").value == 1
    assert key.value("Hex Dword BE").type == RegistryValueType.DWORD_BIG_ENDIAN

    assert key.value("Multi String").value == ["a", "b", "c", "d"]
    assert key.value("Multi String").type == RegistryValueType.MULTI_SZ

    assert key.value("Long Multi String").value == ["a" * 1024, "b" * 1024, "c" * 1024, "d" * 1024]
    assert key.value("Long Multi String").type == RegistryValueType.MULTI_SZ

    assert key.value("Qword").value == 1
    assert key.value("Qword").type == RegistryValueType.QWORD


@pytest.mark.parametrize(
    ("pattern", "has_glob"),
    [
        ("", False),
        ("foo\\bar", False),
        ("foo\\*\\", True),
        ("\\?\\foo", True),
        ("foo\\[", True),
        ("foo\\]", False),
    ],
)
def test_has_glob_magic(pattern: str, has_glob: bool) -> None:
    assert has_glob_magic(pattern) is has_glob


@pytest.mark.parametrize(
    ("pattern", "part1", "part2"),
    [
        ("", "", ""),
        ("foo\\bar", "foo\\bar", ""),
        ("\\foo\\bar", "\\foo\\bar", ""),
        ("foo\\bar\\*\\", "foo\\bar\\", "*\\"),
        ("foo\\bar\\*\\bla", "foo\\bar\\", "*\\bla"),
        ("foo\\bar\\?\\bla", "foo\\bar\\", "?\\bla"),
        ("foo\\bar\\[a-z]\\bla", "foo\\bar\\", "[a-z]\\bla"),
        ("foo\\bar\\b*z\\bla", "foo\\bar\\", "b*z\\bla"),
        ("foo\\bar\\b?z\\bla", "foo\\bar\\", "b?z\\bla"),
        ("foo\\bar\\b[a-z]z\\bla", "foo\\bar\\", "b[a-z]z\\bla"),
        ("*\\foo\\bar\\bla", "", "*\\foo\\bar\\bla"),
        ("?\\foo\\bar\\bla", "", "?\\foo\\bar\\bla"),
        ("[a-z]\\foo\\bar\\bla", "", "[a-z]\\foo\\bar\\bla"),
        ("f*r\\foo\\bar\\bla", "", "f*r\\foo\\bar\\bla"),
        ("f?r\\foo\\bar\\bla", "", "f?r\\foo\\bar\\bla"),
        ("f[a-z]r\\foo\\bar\\bla", "", "f[a-z]r\\foo\\bar\\bla"),
        ("\\*\\foo\\bar\\bla", "\\", "*\\foo\\bar\\bla"),
        ("foo\\bar\\bla\\*", "foo\\bar\\bla\\", "*"),
        ("foo\\bar\\bla\\?", "foo\\bar\\bla\\", "?"),
        ("foo\\bar\\bla\\[a-z]", "foo\\bar\\bla\\", "[a-z]"),
        ("foo\\bar\\bla\\b*z", "foo\\bar\\bla\\", "b*z"),
        ("foo\\bar\\bla\\b?z", "foo\\bar\\bla\\", "b?z"),
        ("foo\\bar\\bla\\b[a-z]z", "foo\\bar\\bla\\", "b[a-z]z"),
    ],
)
def test_glob_split(pattern: str, part1: str, part2: str) -> None:
    assert glob_split(pattern) == (part1, part2)


@pytest.fixture
def hive() -> VirtualHive:
    hive = VirtualHive()

    key_paths = [
        "\\some\\path\\to\\foo",
        "\\some\\path\\to\\bar",
        "\\some\\path\\to\\bla",
        "\\some\\path\\bla",
        "\\some\\other\\bla",
        "\\some\\very-long\\path\\to\\bla",
        "\\some\\other-long\\path\\into\\bla",
        "\\other\\path\\to\\bla",
    ]

    for key_path in key_paths:
        key = VirtualKey(hive, key_path)
        hive.map_key(key_path, key)

    return hive


@pytest.fixture
def hivecollection(hive: RegistryHive) -> HiveCollection:
    return HiveCollection([hive])


@pytest.fixture
def key_collection(hivecollection: HiveCollection) -> KeyCollection:
    return hivecollection.key("\\")


@pytest.mark.parametrize(
    ("key_path", "key_name"),
    [
        ("\\", "VROOT"),
        ("\\some\\path\\to\\foo", "foo"),
        ("\\some\\nonexisting\\meh", RegistryKeyNotFoundError),
        ("some\\other\\bla\\", "bla"),
    ],
)
def test_registry_key_get(hive: RegistryHive, key_path: str, key_name: str | RegistryKeyNotFoundError) -> None:
    key = hive.key("\\")

    if key_name is RegistryKeyNotFoundError:
        with pytest.raises(key_name):
            key.get(key_path)
    else:
        assert key.get(key_path).name == key_name


@pytest.mark.parametrize(
    ("key_path", "key_name"),
    [
        ("\\", "VROOT"),
        ("\\some\\path\\to\\foo", "foo"),
        ("\\some\\nonexisting\\meh", RegistryKeyNotFoundError),
        ("some\\other\\bla\\", "bla"),
    ],
)
def test_key_collection_get(
    key_collection: KeyCollection,
    key_path: str,
    key_name: str | RegistryKeyNotFoundError,
) -> None:
    if key_name is RegistryKeyNotFoundError:
        with pytest.raises(key_name):
            key_collection.get(key_path)
    else:
        assert key_collection.get(key_path).name == key_name


@pytest.mark.parametrize(
    ("key_path", "key_names"),
    [
        ("\\some\\path\\to\\foo", ["foo"]),
        ("\\non\\existing\\key", []),
    ],
)
def test_glob_ext0(key_collection: KeyCollection, key_path: str, key_names: list[str]) -> None:
    key_collections = glob_ext0(key_collection, key_path)

    collection_names = [key_collection.name for key_collection in key_collections]

    assert collection_names == key_names


@pytest.mark.parametrize(
    ("pattern", "key_names"),
    [
        (
            "*",
            ["foo", "bar", "bla"],
        ),
        (
            "b*",
            ["bar", "bla"],
        ),
        ("z*z", []),
        (
            "F*",
            ["foo"],
        ),
    ],
)
def test_glob_ext1(hivecollection: HiveCollection, pattern: str, key_names: list[str]) -> None:
    key_collection = hivecollection.key("\\some\\path\\to\\")
    key_collections = glob_ext1(key_collection, pattern)

    collection_names = [key_collection.name for key_collection in key_collections]

    assert sorted(collection_names) == sorted(key_names)


@pytest.mark.parametrize(
    ("pattern", "key_paths"),
    [
        (
            "\\some\\path\\to\\foo",
            [
                "\\some\\path\\to\\foo",
            ],
        ),
        (
            "\\some\\path\\to\\foo\\",
            [
                "\\some\\path\\to\\foo",
            ],
        ),
        (
            "some\\path\\to\\foo",
            [
                "\\some\\path\\to\\foo",
            ],
        ),
        (
            "\\some\\path\\to\\*",
            [
                "\\some\\path\\to\\foo",
                "\\some\\path\\to\\bar",
                "\\some\\path\\to\\bla",
            ],
        ),
        (
            "\\some\\path\\to\\*\\",
            [
                "\\some\\path\\to\\foo",
                "\\some\\path\\to\\bar",
                "\\some\\path\\to\\bla",
            ],
        ),
        (
            "\\*\\path\\to\\bla",
            [
                "\\some\\path\\to\\bla",
                "\\other\\path\\to\\bla",
            ],
        ),
        (
            "*\\path\\to\\bla",
            [
                "\\some\\path\\to\\bla",
                "\\other\\path\\to\\bla",
            ],
        ),
        (
            "\\some\\*\\bla",
            [
                "\\some\\path\\bla",
                "\\some\\other\\bla",
            ],
        ),
        (
            "\\some\\path\\to\\z*",
            [],
        ),
        (
            "\\*\\path\\to\\zap",
            [],
        ),
        (
            "\\some\\non-existing\\path\\to\\foo",
            [],
        ),
        (
            "\\some\\*\\path\\*\\bla",
            [
                "\\some\\very-long\\path\\to\\bla",
                "\\some\\other-long\\path\\into\\bla",
            ],
        ),
    ],
)
def test_glob_ext(key_collection: KeyCollection, pattern: str, key_paths: list[str]) -> None:
    key_collections = glob_ext(key_collection, pattern)
    collection_paths = [key_collection.path for key_collection in key_collections]

    assert sorted(collection_paths) == sorted(key_paths)


@pytest.mark.parametrize(
    ("input", "expected_name", "expected_value"),
    [
        (c_regf.REG_NONE, "NONE", 0),
        (c_regf.REG_SZ, "SZ", 1),
        (c_regf.REG_EXPAND_SZ, "EXPAND_SZ", 2),
        (c_regf.REG_BINARY, "BINARY", 3),
        (c_regf.REG_DWORD, "DWORD", 4),
        (c_regf.REG_DWORD_BIG_ENDIAN, "DWORD_BIG_ENDIAN", 5),
        (c_regf.REG_LINK, "LINK", 6),
        (c_regf.REG_MULTI_SZ, "MULTI_SZ", 7),
        (c_regf.REG_RESOURCE_LIST, "RESOURCE_LIST", 8),
        (c_regf.REG_FULL_RESOURCE_DESCRIPTOR, "FULL_RESOURCE_DESCRIPTOR", 9),
        (c_regf.REG_RESOURCE_REQUIREMENTS_LIST, "RESOURCE_REQUIREMENTS_LIST", 10),
        (c_regf.REG_QWORD, "QWORD", 11),
        (1337, "1337", 1337),
    ],
)
def test_registry_value_type_enum(input: int, expected_name: str | None, expected_value: int) -> None:
    """Test if registry value types are not parsed strictly within the Enum."""
    regf_value = RegistryValueType(input)
    assert regf_value == expected_value
    assert regf_value.name == expected_name



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.configutil import Env
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export

if TYPE_CHECKING:
    from collections.abc import Iterator

EnvironmentFileRecord = TargetRecordDescriptor(
    "application/other/file/environment",
    [
        ("datetime", "ts_mtime"),
        ("string", "key"),
        ("string", "value"),
        ("string", "comment"),
        ("path", "path"),
    ],
)


class EnvironmentFilePlugin(Plugin):
    """Environment file plugin."""

    def check_compatible(self) -> None:
        # `--env-path` is required at runtime
        pass

    @export(record=EnvironmentFileRecord)
    @arg("--env-path", required=True, help="path to scan environment files in")
    @arg("--extension", default="env", help="extension of files to scan")
    def envfile(self, env_path: str, extension: str = "env") -> Iterator[EnvironmentFileRecord]:
        """Yield environment variables found in ``.env`` files at the provided path."""

        if not env_path:
            self.target.log.error("No ``--path`` provided!")
            return

        if not (path := self.target.fs.path(env_path)).exists():
            self.target.log.error("Provided path %s does not exist!", path)
            return

        for file in path.glob("**/*." + extension):
            if not file.is_file():
                continue

            mtime = file.lstat().st_mtime

            with file.open("r") as fh:
                parser = Env(comments=True)
                parser.read_file(fh)

                for key, (value, comment) in parser.parsed_data.items():
                    yield EnvironmentFileRecord(
                        ts_mtime=mtime,
                        key=key,
                        value=value,
                        comment=comment,
                        path=file,
                        _target=self.target,
                    )



from __future__ import annotations

from io import BytesIO

import pytest
from dissect.cstruct import cstruct

from dissect.target.helpers.protobuf import ProtobufVarint, decode_varint, encode_varint


@pytest.mark.parametrize(
    ("input", "expected_output"),
    [
        (b"\xd2\x85\xd8\xcc\x04", 1234567890),
        (b"\xd2\x85\xd8\xcc\x04\x01\x02\x03", 1234567890),
    ],
)
def test_protobuf_varint_decode(input: bytes, expected_output: int) -> None:
    assert decode_varint(BytesIO(input)) == expected_output


@pytest.mark.parametrize(
    ("input", "expected_output"),
    [
        (1234567890, b"\xd2\x85\xd8\xcc\x04"),
        (pow(2, 128), b"\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x04"),
    ],
)
def test_protobuf_varint_encode(input: int, expected_output: bytes) -> None:
    assert encode_varint(input) == expected_output


def test_protobuf_varint_cstruct() -> None:
    struct_def = """
    struct foo {
        uint32 foo;
        varint size;
        char   bar[size];
    };
    """
    cs = cstruct(endian=">")
    cs.add_custom_type("varint", ProtobufVarint, size=None, alignment=1, signed=False)
    cs.load(struct_def, compiled=False)

    aaa = b"a" * 123456
    buf = b"\x00\x00\x00\x01\xc0\xc4\x07" + aaa
    foo = cs.foo(buf + b"\x01\x02\x03")
    assert foo.foo == 1
    assert foo.size == 123456
    assert foo.bar == aaa
    assert foo.dumps() == buf

    assert cs.varint[2](b"\x80\x01\x80\x02") == [128, 256]
    assert cs.varint[2].dumps([128, 256]) == b"\x80\x01\x80\x02"



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.loader import Loader
from dissect.target.plugins.os.default._os import DefaultOSPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


class DirectLoader(Loader):
    def __init__(self, paths: list[str | Path]):
        self.paths = [(Path(path) if not isinstance(path, Path) else path).resolve() for path in paths]

    @staticmethod
    def detect(path: Path) -> bool:
        return False

    def map(self, target: Target) -> None:
        vfs = VirtualFilesystem()
        for path in self.paths:
            if path.is_file():
                vfs.map_file(str(path), str(path))
            elif path.is_dir():
                vfs.map_dir(str(path), str(path))

        target.filesystems.add(vfs)
        target._os_plugin = DefaultOSPlugin



from __future__ import annotations

from contextlib import AbstractContextManager
from typing import TYPE_CHECKING

from dissect.target.exceptions import Error
from dissect.target.helpers.nfs.nfs3 import MountOK, MountProc, MountStat
from dissect.target.helpers.nfs.serializer import MountResultDeserializer
from dissect.target.helpers.sunrpc.client import AbstractClient as SunRpcAbstractClient
from dissect.target.helpers.sunrpc.client import Client as SunRpcClient
from dissect.target.helpers.sunrpc.client import LocalPortPolicy, auth_null
from dissect.target.helpers.sunrpc.serializer import StringSerializer

if TYPE_CHECKING:
    from types import TracebackType

    from typing_extensions import Self


class MountError(Error):
    def __init__(self, message: str, mount_stat: MountStat):
        super().__init__(message)
        self.message = message
        self.mount_stat = mount_stat

    def __str__(self) -> str:
        return f"{self.__class__.__name__} (Mount stat: {self.mount_stat}): {self.message}"


class Client(AbstractContextManager):
    DIR_COUNT = 4096  # See https://datatracker.ietf.org/doc/html/rfc1813#section-3.3.17
    MAX_COUNT = 32768
    READ_CHUNK_SIZE = 1024 * 1024

    def __init__(self, rpc_client: SunRpcAbstractClient):
        self._rpc_client = rpc_client

    def __enter__(self) -> Self:
        """Return ``self`` upon entering the runtime context."""
        return self  # type: Necessary for type checker

    def __exit__(self, _: type[BaseException] | None, __: BaseException | None, ___: TracebackType | None) -> bool:
        self.close()
        return False  # Reraise exceptions

    @classmethod
    def connect(
        cls,
        hostname: str,
        port: int,
        local_port: int | LocalPortPolicy = 0,
        timeout_in_seconds: float | None = 5.0,
    ) -> Client:
        """Utility function to setup a connection to a NFS Mount Server

        Args:
            hostname: The remote hostname.
            port: The remote port.
            local_port: The local port to bind to.
                If equal to ``LocalPortPolicy.PRIVILEGED`` or -1, bind to the first free privileged port.
                If equal to ``LocalPortPolicy.ANY`` or 0, bind to any free port.
                Otherwise, bind to the specified port.
            timeout_in_seconds: The timeout for making the connection.
        """
        rpc_client = SunRpcClient.connect(hostname, port, auth_null(), local_port, timeout_in_seconds)
        return Client(rpc_client)

    def close(self) -> None:
        self._rpc_client.close()

    def mount(self, remote_path: str) -> MountOK:
        result = self._rpc_client.call(MountProc, remote_path, StringSerializer(), MountResultDeserializer())
        if isinstance(result, MountStat):
            raise MountError("Failed to mount nfs share", result)

        return result



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.filesystems.ntfs import NtfsFilesystem
from dissect.target.plugins.filesystem.icat import ICatPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_unsupported_fs(target_win: Target) -> None:
    # must not fail if sysvol mounted FS is not supported
    assert target_win.icat(1, 0, "") is None


def test_ntfs_fs(target_win: Target) -> None:
    mft_data_file = absolute_path("_data/plugins/filesystem/ntfs/mft/mft.raw")
    filesystem = NtfsFilesystem(mft=mft_data_file.open("rb"))
    target_win.filesystems = [filesystem]
    target_win.add_plugin(ICatPlugin)

    assert target_win.icat(1, 0, "") is None



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from dissect.esedb.exceptions import Error
from dissect.esedb.tools import ual

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target

ClientAccessRecord = TargetRecordDescriptor(
    "filesystem/windows/ual/client_access",
    [
        ("datetime", "last_access_date"),
        ("datetime", "access_date"),
        ("datetime", "insert_date"),
        ("net.ipaddress", "address"),
        ("string", "authenticated_user"),
        ("string", "client_name"),
        ("varint", "access_count"),
        ("varint", "total_access_count"),
        ("string", "tenant_id"),
        ("string", "role_guid"),
        ("string", "role_name"),
        ("string", "product_name"),
        ("string", "path"),
    ],
)

RoleAccessRecord = TargetRecordDescriptor(
    "filesystem/windows/ual/role_access",
    [
        ("datetime", "first_seen_date"),
        ("datetime", "last_seen_date"),
        ("string", "role_guid"),
        ("string", "role_name"),
        ("string", "product_name"),
        ("string", "path"),
    ],
)

VirtualMachineRecord = TargetRecordDescriptor(
    "filesystem/windows/ual/virtual_machines",
    [
        ("datetime", "creation_time"),
        ("datetime", "last_seen_active_date"),
        ("string", "vm_guid"),
        ("string", "bios_guid"),
        ("string", "serial_number"),
        ("string", "path"),
    ],
)

DomainSeenRecord = TargetRecordDescriptor(
    "filesystem/windows/ual/virtual_machines",
    [
        ("datetime", "last_seen_date"),
        ("net.ipaddress", "address"),
        ("string", "resolved_hostname"),
        ("string", "path"),
    ],
)

SystemIdentityRecord = TargetRecordDescriptor(
    "filesystem/windows/ual/system_identity",
    [
        ("datetime", "creation_time"),
        ("varint", "physical_processor_count"),
        ("varint", "cores_per_physical_processor"),
        ("varint", "logical_processors_per_physical_processor"),
        ("varint", "maximum_memory"),
        ("varint", "os_major_version"),
        ("varint", "os_minor_version"),
        ("varint", "os_build"),
        ("varint", "os_platform_id"),
        ("varint", "service_pack_major_version"),
        ("varint", "service_pack_minor_version"),
        ("varint", "os_suite_mask"),
        ("varint", "os_product_type"),
        ("varint", "os_current_time_zone"),
        ("varint", "os_daylight_in_effect"),
        ("string", "system_manufacturer"),
        ("string", "system_product_name"),
        ("string", "system_sm_bios_uuid"),
        ("string", "system_serial_number"),
        ("string", "system_dns_hostname"),
        ("string", "system_domain_name"),
        ("string", "os_serial_number"),
        ("string", "os_country_code"),
        ("string", "os_last_boot_up_time"),
        ("string", "path"),
    ],
)

FIELD_NAME_MAP = {
    "Address": "address",
    "AuthenticatedUserName": "authenticated_user",
    "BIOSGuid": "bios_guid",
    "ClientName": "client_name",
    "CoresPerPhysicalProcessor": "cores_per_physical_processor",
    "CreationTime": "creation_time",
    "FirstSeen": "first_seen_date",
    "HostName": "resolved_hostname",
    "InsertDate": "insert_date",
    "LastAccess": "last_access_date",
    "LastSeen": "last_seen_date",
    "LastSeenActive": "last_seen_active_date",
    "LogicalProcessorsPerPhysicalProcessor": "logical_processors_per_physical_processor",
    "MaximumMemory": "maximum_memory",
    "OSBuildNumber": "os_build",
    "OSCountryCode": "os_country_code",
    "OSCurrentTimeZone": "os_current_time_zone",
    "OSDaylightInEffect": "os_daylight_in_effect",
    "OSLastBootUpTime": "os_last_boot_up_time",
    "OSMajor": "os_major_version",
    "OSMinor": "os_minor_version",
    "OSPlatformId": "os_platform_id",
    "OSProductType": "os_product_type",
    "OSSerialNumber": "os_serial_number",
    "OSSuiteMask": "os_suite_mask",
    "PhysicalProcessorCount": "physical_processor_count",
    "ProductName": "product_name",
    "RoleGuid": "role_guid",
    "RoleName": "role_name",
    "SerialNumber": "serial_number",
    "ServicePackMajor": "service_pack_major_version",
    "ServicePackMinor": "service_pack_minor_version",
    "SystemDNSHostName": "system_dns_hostname",
    "SystemDomainName": "system_domain_name",
    "SystemManufacturer": "system_manufacturer",
    "SystemProductName": "system_product_name",
    "SystemSMBIOSUUID": "system_sm_bios_uuid",
    "SystemSerialNumber": "system_serial_number",
    "TenantId": "tenant_id",
    "TotalAccesses": "total_access_count",
    "VmGuid": "vm_guid",
}


class UalPlugin(Plugin):
    """Return all available User Access Log information.

    User Access Logging (UAL) is a logging system that aggregates client usage data by role and products on a local
    server. It helps Windows server administrators to quantify requests from client computers for roles and services on
    a local server.

    References:
        - https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-server-2012-R2-and-2012/hh849634%28v=ws.11%29
    """

    __namespace__ = "ual"

    LOG_DB_GLOB = "%windir%/System32/LogFiles/Sum/*.mdb"

    IDENTITY_DB_FILENAME = "SystemIdentity.mdb"
    IDENTITY_DB_PATH = f"%windir%/System32/LogFiles/Sum/{IDENTITY_DB_FILENAME}"

    def __init__(self, target: Target):
        super().__init__(target)

        self.mdb_paths = self.find_mdb_files()

        self.role_guid_map = {}
        self.identity_db_parser = None
        self.populate_role_guid_map()

    def check_compatible(self) -> None:
        if not any(path.exists() for path in self.mdb_paths):
            raise UnsupportedPluginError("No MDB files found")

    def find_mdb_files(self) -> list[Path]:
        base, _, glob = self.LOG_DB_GLOB.partition("*")

        return [
            path
            for path in self.target.resolve(base).glob(f"*{glob}")
            if path.exists() and path.name != self.IDENTITY_DB_FILENAME
        ]

    def populate_role_guid_map(self) -> None:
        identity_db = self.target.resolve(self.IDENTITY_DB_PATH)
        if not identity_db.exists():
            return

        fh = identity_db.open()
        try:
            self.identity_db_parser = ual.UAL(fh)
        except Error as e:
            self.target.log.warning("Error opening UAL SystemIdentity.mdb database")
            self.target.log.debug("", exc_info=e)
            return

        self.target.log.debug("SystemIdentity.mdb DB loaded")

        for record in self.identity_db_parser.get_table_records("ROLE_IDS"):
            self.role_guid_map[record.get("RoleGuid")] = {
                "product_name": record.get("ProductName"),
                "role_name": record.get("RoleName"),
            }

    def read_table_records(self, table_name: str) -> Iterator[tuple[Path, dict[str, Any]]]:
        for mdb_path in self.mdb_paths:
            fh = mdb_path.open()
            try:
                parser = ual.UAL(fh)
            except Error as e:
                self.target.log.warning("Error opening database: %s", mdb_path)
                self.target.log.debug("", exc_info=e)
                continue

            for table_record in parser.get_table_records(table_name):
                values = {FIELD_NAME_MAP.get(key, key): value for key, value in table_record.items()}
                yield mdb_path, values

    @export(record=ClientAccessRecord)
    def client_access(self) -> Iterator[ClientAccessRecord]:
        """Return client access data within the User Access Logs."""
        for path, client_record in self.read_table_records("CLIENTS"):
            common_values = {k: v for k, v in client_record.items() if k != "activity_counts"}
            role_guid_data = self.role_guid_map.get(common_values.get("role_guid"), {})

            for access_date, access_count in client_record.get("activity_counts", []):
                yield ClientAccessRecord(
                    role_name=role_guid_data.get("role_name"),
                    product_name=role_guid_data.get("product_name"),
                    access_date=access_date,
                    access_count=access_count,
                    path=path,
                    _target=self.target,
                    **common_values,
                )

    @export(record=RoleAccessRecord)
    def role_access(self) -> Iterator[RoleAccessRecord]:
        """Return role access data within the User Access Logs."""
        for path, record in self.read_table_records("ROLE_ACCESS"):
            role_guid_data = self.role_guid_map.get(record.get("role_guid"), {})
            yield RoleAccessRecord(
                role_name=role_guid_data.get("role_name"),
                product_name=role_guid_data.get("product_name"),
                path=path,
                _target=self.target,
                **record,
            )

    @export(record=VirtualMachineRecord)
    def virtual_machines(self) -> Iterator[VirtualMachineRecord]:
        """Return virtual machine data within the User Access Logs."""
        for path, record in self.read_table_records("VIRTUALMACHINES"):
            yield VirtualMachineRecord(path=path, _target=self.target, **record)

    @export(record=DomainSeenRecord)
    def domains_seen(self) -> Iterator[DomainSeenRecord]:
        """Return DNS data within the User Access Logs."""
        for path, record in self.read_table_records("DNS"):
            yield DomainSeenRecord(path=path, _target=self.target, **record)

    @export(record=SystemIdentityRecord)
    def system_identities(self) -> Iterator[SystemIdentityRecord]:
        """Return system identity data within the User Access Logs."""
        if not self.identity_db_parser:
            return
        for record in self.identity_db_parser.get_table_records("SYSTEM_IDENTITY"):
            values = {FIELD_NAME_MAP.get(key, key): value for key, value in record.items()}
            yield SystemIdentityRecord(
                path=self.target.fs.path(self.IDENTITY_DB_PATH),
                _target=self.target,
                **values,
            )
